#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pytest
import numpy as np
from math import log
import idaes.logger as idaeslog
from pyomo.environ import (
    ConcreteModel,
    Constraint,
    value,
    Var,
    units as pyunits,
    assert_optimal_termination,
    TransformationFactory,
)
from pyomo.network import Arc
from pyomo.util.check_units import assert_units_consistent
from pyomo.network import Port
from idaes.core import (
    FlowsheetBlock,
    MaterialBalanceType,
    MomentumBalanceType,
    ControlVolume0DBlock,
)
from watertap.property_models.multicomp_aq_sol_prop_pack import (
    MCASParameterBlock,
    ActivityCoefficientModel,
    DensityCalculation,
    MCASStateBlock,
)
from watertap.unit_models.nanofiltration_DSPMDE_0D import (
    NanofiltrationDSPMDE0D,
    MassTransferCoefficient,
    ConcentrationPolarizationType,
)
from watertap.core.util.initialization import check_dof

from idaes.core.solvers import get_solver
from idaes.core.util.model_statistics import (
    number_variables,
    number_total_constraints,
    number_unused_variables,
)
from idaes.core.util.testing import initialization_tester
from idaes.core.util.exceptions import ConfigurationError
from idaes.core.util.scaling import (
    calculate_scaling_factors,
    unscaled_variables_generator,
    badly_scaled_var_generator,
)
from idaes.core.util.initialization import propagate_state
from idaes.models.unit_models import (
    Feed,
)
import idaes.logger as idaeslog

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()


@pytest.mark.unit
def test_config_with_CP():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = MCASParameterBlock(
        solute_list=["Ca_2+", "SO4_2-", "Na_+", "Cl_-", "Mg_2+"],
        charge={"Ca_2+": 2, "SO4_2-": -2, "Na_+": 1, "Cl_-": -1, "Mg_2+": 2},
        diffusivity_data={
            ("Liq", "Ca_2+"): 9.2e-10,
            ("Liq", "SO4_2-"): 1.06e-09,
            ("Liq", "Mg_2+"): 7.06e-10,
            ("Liq", "Na_+"): 1.33e-09,
            ("Liq", "Cl_-"): 2.03e-09,
        },
        mw_data={
            "Ca_2+": 40e-3,
            "SO4_2-": 97e-3,
            "Na_+": 23e-3,
            "Cl_-": 35e-3,
            "Mg_2+": 24e-3,
        },
    )
    m.fs.unit = NanofiltrationDSPMDE0D(property_package=m.fs.properties)

    # check unit config arguments
    assert len(m.fs.unit.config) == 9

    assert not m.fs.unit.config.dynamic
    assert not m.fs.unit.config.has_holdup
    assert m.fs.unit.config.material_balance_type == MaterialBalanceType.useDefault
    assert m.fs.unit.config.momentum_balance_type == MomentumBalanceType.pressureTotal
    assert not m.fs.unit.config.has_pressure_change
    assert not hasattr(m.fs.unit.config, "energy_balance_type")
    assert m.fs.unit.config.property_package is m.fs.properties
    assert (
        m.fs.unit.config.property_package.config.activity_coefficient_model
        == ActivityCoefficientModel.ideal
    )
    assert (
        m.fs.unit.config.property_package.config.density_calculation
        == DensityCalculation.constant
    )
    assert (
        m.fs.unit.config.mass_transfer_coefficient
        == MassTransferCoefficient.spiral_wound
    )
    assert (
        m.fs.unit.config.concentration_polarization_type
        == ConcentrationPolarizationType.calculated
    )
    assert hasattr(m.fs.unit.config.property_package, "solute_set")
    assert hasattr(m.fs.unit.config.property_package, "ion_set")
    assert hasattr(m.fs.unit.config.property_package, "cation_set")
    assert hasattr(m.fs.unit.config.property_package, "anion_set")
    assert_units_consistent(m)


@pytest.mark.unit
def test_config_without_CP():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = MCASParameterBlock(
        solute_list=["Ca_2+", "SO4_2-", "Na_+", "Cl_-", "Mg_2+"],
        charge={"Ca_2+": 2, "SO4_2-": -2, "Na_+": 1, "Cl_-": -1, "Mg_2+": 2},
        diffusivity_data={
            ("Liq", "Ca_2+"): 9.2e-10,
            ("Liq", "SO4_2-"): 1.06e-09,
            ("Liq", "Mg_2+"): 7.06e-10,
            ("Liq", "Na_+"): 1.33e-09,
            ("Liq", "Cl_-"): 2.03e-09,
        },
        mw_data={
            "Ca_2+": 40e-3,
            "SO4_2-": 97e-3,
            "Na_+": 23e-3,
            "Cl_-": 35e-3,
            "Mg_2+": 24e-3,
        },
    )
    m.fs.unit = NanofiltrationDSPMDE0D(
        property_package=m.fs.properties,
        mass_transfer_coefficient=MassTransferCoefficient.none,
        concentration_polarization_type=ConcentrationPolarizationType.none,
    )

    # check unit config arguments
    assert len(m.fs.unit.config) == 9

    assert not m.fs.unit.config.dynamic
    assert not m.fs.unit.config.has_holdup
    assert m.fs.unit.config.material_balance_type == MaterialBalanceType.useDefault
    assert m.fs.unit.config.momentum_balance_type == MomentumBalanceType.pressureTotal
    assert not m.fs.unit.config.has_pressure_change
    assert not hasattr(m.fs.unit.config, "energy_balance_type")
    assert m.fs.unit.config.property_package is m.fs.properties
    assert (
        m.fs.unit.config.property_package.config.activity_coefficient_model
        == ActivityCoefficientModel.ideal
    )
    assert (
        m.fs.unit.config.property_package.config.density_calculation
        == DensityCalculation.constant
    )
    assert m.fs.unit.config.mass_transfer_coefficient == MassTransferCoefficient.none
    assert (
        m.fs.unit.config.concentration_polarization_type
        == ConcentrationPolarizationType.none
    )
    assert hasattr(m.fs.unit.config.property_package, "solute_set")
    assert hasattr(m.fs.unit.config.property_package, "ion_set")
    assert hasattr(m.fs.unit.config.property_package, "cation_set")
    assert hasattr(m.fs.unit.config.property_package, "anion_set")
    assert_units_consistent(m)


class TestNanoFiltration_with_CP_5ions:
    @pytest.fixture(scope="class")
    def NF_frame(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.properties = MCASParameterBlock(
            solute_list=["Ca_2+", "SO4_2-", "Mg_2+", "Na_+", "Cl_-"],
            diffusivity_data={
                ("Liq", "Ca_2+"): 9.2e-10,
                ("Liq", "SO4_2-"): 1.06e-09,
                ("Liq", "Mg_2+"): 7.06e-10,
                ("Liq", "Na_+"): 1.33e-09,
                ("Liq", "Cl_-"): 2.03e-09,
            },
            mw_data={
                "H2O": 0.018,
                "Ca_2+": 0.04,
                "Mg_2+": 0.024,
                "SO4_2-": 0.096,
                "Na_+": 0.023,
                "Cl_-": 0.035,
            },
            stokes_radius_data={
                "Ca_2+": 3.09e-10,
                "Mg_2+": 3.47e-10,
                "SO4_2-": 2.3e-10,
                "Cl_-": 1.21e-10,
                "Na_+": 1.84e-10,
            },
            charge={"Ca_2+": 2, "Mg_2+": 2, "SO4_2-": -2, "Na_+": 1, "Cl_-": -1},
            activity_coefficient_model=ActivityCoefficientModel.davies,
            density_calculation=DensityCalculation.constant,
        )

        m.fs.unit = NanofiltrationDSPMDE0D(property_package=m.fs.properties)
        b = m.fs.unit
        mass_flow_in = 1 * pyunits.kg / pyunits.s
        feed_mass_frac = {
            "Ca_2+": 382e-6,
            "Mg_2+": 1394e-6,
            "SO4_2-": 2136e-6,
            "Cl_-": 20101.6e-6,
            "Na_+": 11122e-6,
        }

        # Fix mole flow rates of each ion and water
        for ion, x in feed_mass_frac.items():
            mol_comp_flow = (
                x
                * pyunits.kg
                / pyunits.kg
                * mass_flow_in
                / m.fs.unit.feed_side.properties_in[0].mw_comp[ion]
            )
            m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", ion].fix(mol_comp_flow)
        H2O_mass_frac = 1 - sum(x for x in feed_mass_frac.values())
        H2O_mol_comp_flow = (
            H2O_mass_frac
            * pyunits.kg
            / pyunits.kg
            * mass_flow_in
            / m.fs.unit.feed_side.properties_in[0].mw_comp["H2O"]
        )
        m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "H2O"].fix(H2O_mol_comp_flow)

        # Use assert electroneutrality method from property model to ensure the ion concentrations provided
        # obey electroneutrality condition
        m.fs.unit.feed_side.properties_in[0].assert_electroneutrality(
            defined_state=True,
            adjust_by_ion="Cl_-",
            get_property="mass_frac_phase_comp",
        )

        # Fix other inlet state variables
        m.fs.unit.inlet.temperature[0].fix(298.15)
        m.fs.unit.inlet.pressure[0].fix(4e5)

        # Fix the membrane variables that are usually fixed for the DSPM-DE model
        m.fs.unit.radius_pore.fix(0.5e-9)
        m.fs.unit.membrane_thickness_effective.fix(1.33e-6)
        m.fs.unit.membrane_charge_density.fix(-27)
        m.fs.unit.dielectric_constant_pore.fix(41.3)

        # Fix final permeate pressure to be ~atmospheric
        m.fs.unit.mixed_permeate[0].pressure.fix(101325)

        m.fs.unit.spacer_porosity.fix(0.85)
        m.fs.unit.channel_height.fix(5e-4)
        m.fs.unit.velocity[0, 0].fix(0.25)
        m.fs.unit.area.fix(50)
        # Fix additional variables for calculating mass transfer coefficient with spiral wound correlation
        m.fs.unit.spacer_mixing_efficiency.fix()
        m.fs.unit.spacer_mixing_length.fix()
        return m

    @pytest.mark.unit
    def test_build(self, NF_frame):
        m = NF_frame

        # test ports and variables
        port_lst = ["inlet", "retentate", "permeate"]
        port_vars_lst = ["flow_mol_phase_comp", "pressure", "temperature"]
        for port_str in port_lst:
            assert hasattr(m.fs.unit, port_str)
            port = getattr(m.fs.unit, port_str)
            assert len(port.vars) == 3
            assert isinstance(port, Port)
            for var_str in port_vars_lst:
                assert hasattr(port, var_str)
                var = getattr(port, var_str)
                assert isinstance(var, Var)

        # test unit objects (including parameters, variables, and constraints)
        unit_objs_lst = []
        for obj_str in unit_objs_lst:
            assert hasattr(m.fs.unit, obj_str)

        # test state block objects
        assert isinstance(m.fs.unit.feed_side, ControlVolume0DBlock)
        cv_stateblock_lst = [
            "properties_in",
            "properties_out",
            "properties_interface",
        ]
        # feed side
        for sb_str in cv_stateblock_lst:
            sb = getattr(m.fs.unit.feed_side, sb_str)
            assert isinstance(sb, MCASStateBlock)
        # test objects added to control volume
        cv_objs_type_dict = {"eq_feed_interface_isothermal": Constraint}
        for obj_str, obj_type in cv_objs_type_dict.items():
            obj = getattr(m.fs.unit.feed_side, obj_str)
            assert isinstance(obj, obj_type)
        # permeate side
        assert isinstance(m.fs.unit.permeate_side, MCASStateBlock)
        assert isinstance(m.fs.unit.mixed_permeate, MCASStateBlock)
        # membrane
        assert isinstance(m.fs.unit.pore_entrance, MCASStateBlock)
        assert isinstance(m.fs.unit.pore_exit, MCASStateBlock)

        # test statistics
        assert number_variables(m) == 566
        assert number_total_constraints(m) == 526
        assert number_unused_variables(m) == 11

    @pytest.mark.unit
    def test_dof(self, NF_frame):
        m = NF_frame
        check_dof(m, fail_flag=True)

    @pytest.mark.unit
    def test_calculate_scaling(self, NF_frame):
        m = NF_frame

        m.fs.properties.set_default_scaling(
            "flow_mol_phase_comp", 1e4, index=("Liq", "Ca_2+")
        )
        m.fs.properties.set_default_scaling(
            "flow_mol_phase_comp", 1e3, index=("Liq", "SO4_2-")
        )
        m.fs.properties.set_default_scaling(
            "flow_mol_phase_comp", 1e3, index=("Liq", "Mg_2+")
        )
        m.fs.properties.set_default_scaling(
            "flow_mol_phase_comp", 1e2, index=("Liq", "Cl_-")
        )
        m.fs.properties.set_default_scaling(
            "flow_mol_phase_comp", 1e2, index=("Liq", "Na_+")
        )
        m.fs.properties.set_default_scaling(
            "flow_mol_phase_comp", 1e0, index=("Liq", "H2O")
        )

        calculate_scaling_factors(m)

        # check that all variables have scaling factors
        unscaled_var_list = list(unscaled_variables_generator(m.fs.unit))
        assert len(unscaled_var_list) == 0

    @pytest.mark.component
    def test_initialize(self, NF_frame):
        m = NF_frame
        initialization_tester(m, outlvl=idaeslog.DEBUG)

        badly_scaled_var_lst = list(
            badly_scaled_var_generator(m.fs.unit, small=1e-5, zero=1e-12)
        )
        for var, val in badly_scaled_var_lst:
            print(var.name, val)
        assert len(badly_scaled_var_lst) == 0

    @pytest.mark.component
    def test_solve(self, NF_frame):
        m = NF_frame
        results = solver.solve(m, tee=True)

        # Check for optimal solution
        assert_optimal_termination(results)

    @pytest.mark.component
    def test_conservation(self, NF_frame):
        m = NF_frame
        b = m.fs.unit
        comp_lst = m.fs.properties.solute_set

        flow_mass_inlet = sum(
            b.feed_side.properties_in[0].flow_mass_phase_comp["Liq", j]
            for j in comp_lst
        )
        flow_mass_retentate = sum(
            b.feed_side.properties_out[0].flow_mass_phase_comp["Liq", j]
            for j in comp_lst
        )
        flow_mass_permeate = sum(
            b.mixed_permeate[0].flow_mass_phase_comp["Liq", j] for j in comp_lst
        )

        assert (
            abs(value(flow_mass_inlet - flow_mass_retentate - flow_mass_permeate))
            <= 1e-6
        )

    @pytest.mark.component
    def test_solution(self, NF_frame):
        m = NF_frame

        mole_flux_dict = {
            "Na_+": 0.00029831,
            "Cl_-": 0.00031137,
            "Ca_2+": 1.9289e-06,
            "SO4_2-": 1.95859e-06,
            "Mg_2+": 6.56667e-06,
            "H2O": 0.032671,
        }
        for j, val in mole_flux_dict.items():
            assert pytest.approx(val, rel=1e-3) == value(
                m.fs.unit.flux_mol_phase_comp_avg[0, "Liq", j]
            )

        # TODO: subsequently focus on the segment below during the validation and refinement phase
        intrinsic_rejection_dict = {
            "Na_+": -0.0265173,
            "Cl_-": 0.11538899,
            "Ca_2+": 0.6905496,
            "SO4_2-": 0.8687813,
            "Mg_2+": 0.83309579,
        }
        for j, val in intrinsic_rejection_dict.items():
            assert pytest.approx(val, rel=1e-3) == value(
                m.fs.unit.rejection_intrinsic_phase_comp[0, "Liq", j]
            )

    @pytest.mark.unit
    def test_report(self, NF_frame):
        NF_frame.fs.unit.report()


class TestNanoFiltration_without_CP_5ions:
    @pytest.fixture(scope="class")
    def NF_frame(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.properties = MCASParameterBlock(
            solute_list=["Ca_2+", "SO4_2-", "Mg_2+", "Na_+", "Cl_-"],
            diffusivity_data={
                ("Liq", "Ca_2+"): 9.2e-10,
                ("Liq", "SO4_2-"): 1.06e-09,
                ("Liq", "Mg_2+"): 7.06e-10,
                ("Liq", "Na_+"): 1.33e-09,
                ("Liq", "Cl_-"): 2.03e-09,
            },
            mw_data={
                "H2O": 0.018,
                "Ca_2+": 0.04,
                "Mg_2+": 0.024,
                "SO4_2-": 0.096,
                "Na_+": 0.023,
                "Cl_-": 0.035,
            },
            stokes_radius_data={
                "Ca_2+": 3.09e-10,
                "Mg_2+": 3.47e-10,
                "SO4_2-": 2.3e-10,
                "Cl_-": 1.21e-10,
                "Na_+": 1.84e-10,
            },
            charge={"Ca_2+": 2, "Mg_2+": 2, "SO4_2-": -2, "Na_+": 1, "Cl_-": -1},
            activity_coefficient_model=ActivityCoefficientModel.davies,
            density_calculation=DensityCalculation.constant,
        )

        m.fs.unit = NanofiltrationDSPMDE0D(
            property_package=m.fs.properties,
            mass_transfer_coefficient=MassTransferCoefficient.none,
            concentration_polarization_type=ConcentrationPolarizationType.none,
        )
        b = m.fs.unit
        mass_flow_in = 1 * pyunits.kg / pyunits.s
        feed_mass_frac = {
            "Ca_2+": 382e-6,
            "Mg_2+": 1394e-6,
            "SO4_2-": 2136e-6,
            "Cl_-": 20101.6e-6,
            "Na_+": 11122e-6,
        }

        # Fix mole flow rates of each ion and water
        for ion, x in feed_mass_frac.items():
            mol_comp_flow = (
                x
                * pyunits.kg
                / pyunits.kg
                * mass_flow_in
                / m.fs.unit.feed_side.properties_in[0].mw_comp[ion]
            )
            m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", ion].fix(mol_comp_flow)
        H2O_mass_frac = 1 - sum(x for x in feed_mass_frac.values())
        H2O_mol_comp_flow = (
            H2O_mass_frac
            * pyunits.kg
            / pyunits.kg
            * mass_flow_in
            / m.fs.unit.feed_side.properties_in[0].mw_comp["H2O"]
        )
        m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "H2O"].fix(H2O_mol_comp_flow)

        # Use assert electroneutrality method from property model to ensure the ion concentrations provided
        # obey electroneutrality condition
        m.fs.unit.feed_side.properties_in[0].assert_electroneutrality(
            defined_state=True,
            adjust_by_ion="Cl_-",
            get_property="mass_frac_phase_comp",
        )

        # Fix other inlet state variables
        m.fs.unit.inlet.temperature[0].fix(298.15)
        m.fs.unit.inlet.pressure[0].fix(4e5)

        # Fix the membrane variables that are usually fixed for the DSPM-DE model
        m.fs.unit.radius_pore.fix(0.5e-9)
        m.fs.unit.membrane_thickness_effective.fix(1.33e-6)
        m.fs.unit.membrane_charge_density.fix(-27)
        m.fs.unit.dielectric_constant_pore.fix(41.3)

        # Fix final permeate pressure to be ~atmospheric
        m.fs.unit.mixed_permeate[0].pressure.fix(101325)

        m.fs.unit.spacer_porosity.fix(0.85)
        m.fs.unit.channel_height.fix(5e-4)
        m.fs.unit.velocity[0, 0].fix(0.25)
        m.fs.unit.area.fix(50)

        return m

    @pytest.mark.unit
    def test_build(self, NF_frame):
        m = NF_frame

        # test ports and variables
        port_lst = ["inlet", "retentate", "permeate"]
        port_vars_lst = ["flow_mol_phase_comp", "pressure", "temperature"]
        for port_str in port_lst:
            assert hasattr(m.fs.unit, port_str)
            port = getattr(m.fs.unit, port_str)
            assert len(port.vars) == 3
            assert isinstance(port, Port)
            for var_str in port_vars_lst:
                assert hasattr(port, var_str)
                var = getattr(port, var_str)
                assert isinstance(var, Var)

        # test unit objects (including parameters, variables, and constraints)
        unit_objs_lst = []
        for obj_str in unit_objs_lst:
            assert hasattr(m.fs.unit, obj_str)

        # test state block objects
        assert isinstance(m.fs.unit.feed_side, ControlVolume0DBlock)
        cv_stateblock_lst = [
            "properties_in",
            "properties_out",
            "properties_interface",
        ]
        # feed side
        for sb_str in cv_stateblock_lst:
            sb = getattr(m.fs.unit.feed_side, sb_str)
            assert isinstance(sb, MCASStateBlock)
        # test objects added to control volume
        cv_objs_type_dict = {"eq_feed_interface_isothermal": Constraint}
        for obj_str, obj_type in cv_objs_type_dict.items():
            obj = getattr(m.fs.unit.feed_side, obj_str)
            assert isinstance(obj, obj_type)
        # permeate side
        assert isinstance(m.fs.unit.permeate_side, MCASStateBlock)
        assert isinstance(m.fs.unit.mixed_permeate, MCASStateBlock)
        # membrane
        assert isinstance(m.fs.unit.pore_entrance, MCASStateBlock)
        assert isinstance(m.fs.unit.pore_exit, MCASStateBlock)

        # test statistics
        assert number_variables(m) == 532
        assert number_total_constraints(m) == 494
        assert number_unused_variables(m) == 11

    @pytest.mark.unit
    def test_dof(self, NF_frame):
        m = NF_frame
        check_dof(m, fail_flag=True)

    @pytest.mark.unit
    def test_calculate_scaling(self, NF_frame):
        m = NF_frame

        m.fs.properties.set_default_scaling(
            "flow_mol_phase_comp", 1e4, index=("Liq", "Ca_2+")
        )
        m.fs.properties.set_default_scaling(
            "flow_mol_phase_comp", 1e3, index=("Liq", "SO4_2-")
        )
        m.fs.properties.set_default_scaling(
            "flow_mol_phase_comp", 1e3, index=("Liq", "Mg_2+")
        )
        m.fs.properties.set_default_scaling(
            "flow_mol_phase_comp", 1e2, index=("Liq", "Cl_-")
        )
        m.fs.properties.set_default_scaling(
            "flow_mol_phase_comp", 1e2, index=("Liq", "Na_+")
        )
        m.fs.properties.set_default_scaling(
            "flow_mol_phase_comp", 1e0, index=("Liq", "H2O")
        )

        calculate_scaling_factors(m)

        # check that all variables have scaling factors
        unscaled_var_list = list(unscaled_variables_generator(m.fs.unit))
        assert len(unscaled_var_list) == 0

    @pytest.mark.component
    def test_initialize(self, NF_frame):
        m = NF_frame
        initialization_tester(m, outlvl=idaeslog.DEBUG)

        badly_scaled_var_lst = list(
            badly_scaled_var_generator(m.fs.unit, small=1e-5, zero=1e-12)
        )
        for var, val in badly_scaled_var_lst:
            print(var.name, val)
        assert len(badly_scaled_var_lst) == 0

    @pytest.mark.component
    def test_solve(self, NF_frame):
        m = NF_frame
        results = solver.solve(m, tee=True)

        # Check for optimal solution
        assert_optimal_termination(results)

    @pytest.mark.component
    def test_conservation(self, NF_frame):
        m = NF_frame
        b = m.fs.unit
        comp_lst = m.fs.properties.solute_set

        flow_mass_inlet = sum(
            b.feed_side.properties_in[0].flow_mass_phase_comp["Liq", j]
            for j in comp_lst
        )
        flow_mass_retentate = sum(
            b.feed_side.properties_out[0].flow_mass_phase_comp["Liq", j]
            for j in comp_lst
        )
        flow_mass_permeate = sum(
            b.mixed_permeate[0].flow_mass_phase_comp["Liq", j] for j in comp_lst
        )

        assert (
            abs(value(flow_mass_inlet - flow_mass_retentate - flow_mass_permeate))
            <= 1e-6
        )

    @pytest.mark.component
    def test_solution(self, NF_frame):
        m = NF_frame

        mole_flux_dict = {
            "Na_+": 0.0003574,
            "Cl_-": 0.0003694,
            "Ca_2+": 1.860337e-6,
            "SO4_2-": 1.850918e-6,
            "Mg_2+": 5.972139e-06,
            "H2O": 0.039150,
        }
        for j, val in mole_flux_dict.items():
            assert pytest.approx(val, rel=1e-3) == value(
                m.fs.unit.flux_mol_phase_comp_avg[0, "Liq", j]
            )

        # TODO: subsequently focus on the segment below during the validation and refinement phase
        intrinsic_rejection_dict = {
            "Na_+": -0.017782,
            "Cl_-": 0.1143856,
            "Ca_2+": 0.731772,
            "SO4_2-": 0.8854559,
            "Mg_2+": 0.8584226,
        }
        for j, val in intrinsic_rejection_dict.items():
            assert pytest.approx(val, rel=1e-3) == value(
                m.fs.unit.rejection_intrinsic_phase_comp[0, "Liq", j]
            )

    @pytest.mark.unit
    def test_report(self, NF_frame):
        NF_frame.fs.unit.report()


class TestNanoFiltration_with_CP_2ions:
    @pytest.fixture(scope="class")
    def NF_frame(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.properties = MCASParameterBlock(
            solute_list=["Na_+", "Cl_-"],
            diffusivity_data={("Liq", "Na_+"): 1.33e-09, ("Liq", "Cl_-"): 2.03e-09},
            mw_data={"H2O": 0.018, "Na_+": 0.023, "Cl_-": 0.035},
            stokes_radius_data={"Cl_-": 1.21e-10, "Na_+": 1.84e-10},
            charge={"Na_+": 1, "Cl_-": -1},
            activity_coefficient_model=ActivityCoefficientModel.davies,
            density_calculation=DensityCalculation.constant,
        )

        m.fs.unit = NanofiltrationDSPMDE0D(property_package=m.fs.properties)
        b = m.fs.unit
        mass_flow_in = 1 * pyunits.kg / pyunits.s
        feed_mass_frac = {
            "Cl_-": 20101.6e-6,
            "Na_+": 11122e-6,
        }

        # Fix mole flow rates of each ion and water
        for ion, x in feed_mass_frac.items():
            mol_comp_flow = (
                x
                * pyunits.kg
                / pyunits.kg
                * mass_flow_in
                / m.fs.unit.feed_side.properties_in[0].mw_comp[ion]
            )
            m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", ion].fix(mol_comp_flow)
        H2O_mass_frac = 1 - sum(x for x in feed_mass_frac.values())
        H2O_mol_comp_flow = (
            H2O_mass_frac
            * pyunits.kg
            / pyunits.kg
            * mass_flow_in
            / m.fs.unit.feed_side.properties_in[0].mw_comp["H2O"]
        )
        m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "H2O"].fix(H2O_mol_comp_flow)

        # Use assert electroneutrality method from property model to ensure the ion concentrations provided
        # obey electroneutrality condition
        m.fs.unit.feed_side.properties_in[0].assert_electroneutrality(
            defined_state=True,
            adjust_by_ion="Cl_-",
            get_property="mass_frac_phase_comp",
        )

        # Fix other inlet state variables
        m.fs.unit.inlet.temperature[0].fix(298.15)
        m.fs.unit.inlet.pressure[0].fix(4e5)

        # Fix the membrane variables that are usually fixed for the DSPM-DE model
        m.fs.unit.radius_pore.fix(0.5e-9)
        m.fs.unit.membrane_thickness_effective.fix(1.33e-6)
        m.fs.unit.membrane_charge_density.fix(-27)
        m.fs.unit.dielectric_constant_pore.fix(41.3)

        # Fix final permeate pressure to be ~atmospheric
        m.fs.unit.mixed_permeate[0].pressure.fix(101325)

        m.fs.unit.spacer_porosity.fix(0.85)
        m.fs.unit.channel_height.fix(5e-4)
        m.fs.unit.velocity[0, 0].fix(0.25)
        m.fs.unit.area.fix(50)
        # Fix additional variables for calculating mass transfer coefficient with spiral wound correlation
        m.fs.unit.spacer_mixing_efficiency.fix()
        m.fs.unit.spacer_mixing_length.fix()
        return m

    @pytest.mark.unit
    def test_build(self, NF_frame):
        m = NF_frame

        # test ports and variables
        port_lst = ["inlet", "retentate", "permeate"]
        port_vars_lst = ["flow_mol_phase_comp", "pressure", "temperature"]
        for port_str in port_lst:
            assert hasattr(m.fs.unit, port_str)
            port = getattr(m.fs.unit, port_str)
            assert len(port.vars) == 3
            assert isinstance(port, Port)
            for var_str in port_vars_lst:
                assert hasattr(port, var_str)
                var = getattr(port, var_str)
                assert isinstance(var, Var)

        # test unit objects (including parameters, variables, and constraints)
        unit_objs_lst = []
        for obj_str in unit_objs_lst:
            assert hasattr(m.fs.unit, obj_str)

        # test state block objects
        assert isinstance(m.fs.unit.feed_side, ControlVolume0DBlock)
        cv_stateblock_lst = [
            "properties_in",
            "properties_out",
            "properties_interface",
        ]
        # feed side
        for sb_str in cv_stateblock_lst:
            sb = getattr(m.fs.unit.feed_side, sb_str)
            assert isinstance(sb, MCASStateBlock)
        # test objects added to control volume
        cv_objs_type_dict = {"eq_feed_interface_isothermal": Constraint}
        for obj_str, obj_type in cv_objs_type_dict.items():
            obj = getattr(m.fs.unit.feed_side, obj_str)
            assert isinstance(obj, obj_type)
        # permeate side
        assert isinstance(m.fs.unit.permeate_side, MCASStateBlock)
        assert isinstance(m.fs.unit.mixed_permeate, MCASStateBlock)
        # membrane
        assert isinstance(m.fs.unit.pore_entrance, MCASStateBlock)
        assert isinstance(m.fs.unit.pore_exit, MCASStateBlock)

        # test statistics
        assert number_variables(m) == 320
        assert number_total_constraints(m) == 286
        assert number_unused_variables(m) == 11

    @pytest.mark.unit
    def test_dof(self, NF_frame):
        m = NF_frame
        check_dof(m, fail_flag=True)

    @pytest.mark.unit
    def test_calculate_scaling(self, NF_frame):
        m = NF_frame

        m.fs.properties.set_default_scaling(
            "flow_mol_phase_comp", 1e2, index=("Liq", "Cl_-")
        )
        m.fs.properties.set_default_scaling(
            "flow_mol_phase_comp", 1e2, index=("Liq", "Na_+")
        )
        m.fs.properties.set_default_scaling(
            "flow_mol_phase_comp", 1e0, index=("Liq", "H2O")
        )

        calculate_scaling_factors(m)

        # check that all variables have scaling factors
        unscaled_var_list = list(unscaled_variables_generator(m.fs.unit))
        assert len(unscaled_var_list) == 0

    @pytest.mark.component
    def test_initialize(self, NF_frame):
        m = NF_frame
        initialization_tester(m, outlvl=idaeslog.DEBUG)

        badly_scaled_var_lst = list(
            badly_scaled_var_generator(m.fs.unit, small=1e-5, zero=1e-12)
        )
        for var, val in badly_scaled_var_lst:
            print(var.name, val)
        assert len(badly_scaled_var_lst) == 0

    @pytest.mark.component
    def test_solve(self, NF_frame):
        m = NF_frame
        results = solver.solve(m, tee=True)

        # Check for optimal solution
        assert_optimal_termination(results)

    @pytest.mark.component
    def test_conservation(self, NF_frame):
        m = NF_frame
        b = m.fs.unit
        comp_lst = m.fs.properties.solute_set

        flow_mass_inlet = sum(
            b.feed_side.properties_in[0].flow_mass_phase_comp["Liq", j]
            for j in comp_lst
        )
        flow_mass_retentate = sum(
            b.feed_side.properties_out[0].flow_mass_phase_comp["Liq", j]
            for j in comp_lst
        )
        flow_mass_permeate = sum(
            b.mixed_permeate[0].flow_mass_phase_comp["Liq", j] for j in comp_lst
        )

        assert (
            abs(value(flow_mass_inlet - flow_mass_retentate - flow_mass_permeate))
            <= 1e-6
        )

    @pytest.mark.component
    def test_solution(self, NF_frame):
        m = NF_frame

        mole_flux_dict = {
            "Na_+": 0.0010739,
            "Cl_-": 0.0010739,
            "H2O": 0.129305,
        }
        for j, val in mole_flux_dict.items():
            assert pytest.approx(val, rel=1e-3) == value(
                m.fs.unit.flux_mol_phase_comp_avg[0, "Liq", j]
            )

        # TODO: subsequently focus on the segment below during the validation and refinement phase
        intrinsic_rejection_dict = {
            "Na_+": 0.0964865,
            "Cl_-": 0.0964865,
        }
        for j, val in intrinsic_rejection_dict.items():
            assert pytest.approx(val, rel=1e-3) == value(
                m.fs.unit.rejection_intrinsic_phase_comp[0, "Liq", j]
            )

        flow_mol_out_dict = {
            "Na_+": 0.429868,
            "Cl_-": 0.429868,
        }
        for j, val in flow_mol_out_dict.items():
            assert pytest.approx(val, rel=1e-3) == value(
                m.fs.unit.retentate.flow_mol_phase_comp[0, "Liq", j]
            )

    @pytest.mark.unit
    def test_report(self, NF_frame):
        NF_frame.fs.unit.report()


class TestNanoFiltration_without_CP_2ions:
    @pytest.fixture(scope="class")
    def NF_frame(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.properties = MCASParameterBlock(
            solute_list=["Na_+", "Cl_-"],
            diffusivity_data={("Liq", "Na_+"): 1.33e-09, ("Liq", "Cl_-"): 2.03e-09},
            mw_data={"H2O": 0.018, "Na_+": 0.023, "Cl_-": 0.035},
            stokes_radius_data={"Cl_-": 1.21e-10, "Na_+": 1.84e-10},
            charge={"Na_+": 1, "Cl_-": -1},
            activity_coefficient_model=ActivityCoefficientModel.davies,
            density_calculation=DensityCalculation.constant,
        )

        m.fs.unit = NanofiltrationDSPMDE0D(
            property_package=m.fs.properties,
            mass_transfer_coefficient=MassTransferCoefficient.none,
            concentration_polarization_type=ConcentrationPolarizationType.none,
        )
        b = m.fs.unit
        mass_flow_in = 1 * pyunits.kg / pyunits.s
        feed_mass_frac = {
            "Cl_-": 20101.6e-6,
            "Na_+": 11122e-6,
        }

        # Fix mole flow rates of each ion and water
        for ion, x in feed_mass_frac.items():
            mol_comp_flow = (
                x
                * pyunits.kg
                / pyunits.kg
                * mass_flow_in
                / m.fs.unit.feed_side.properties_in[0].mw_comp[ion]
            )
            m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", ion].fix(mol_comp_flow)
        H2O_mass_frac = 1 - sum(x for x in feed_mass_frac.values())
        H2O_mol_comp_flow = (
            H2O_mass_frac
            * pyunits.kg
            / pyunits.kg
            * mass_flow_in
            / m.fs.unit.feed_side.properties_in[0].mw_comp["H2O"]
        )
        m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "H2O"].fix(H2O_mol_comp_flow)

        # Use assert electroneutrality method from property model to ensure the ion concentrations provided
        # obey electroneutrality condition
        m.fs.unit.feed_side.properties_in[0].assert_electroneutrality(
            defined_state=True,
            adjust_by_ion="Cl_-",
            get_property="mass_frac_phase_comp",
        )

        # Fix other inlet state variables
        m.fs.unit.inlet.temperature[0].fix(298.15)
        m.fs.unit.inlet.pressure[0].fix(4e5)

        # Fix the membrane variables that are usually fixed for the DSPM-DE model
        m.fs.unit.radius_pore.fix(0.5e-9)
        m.fs.unit.membrane_thickness_effective.fix(1.33e-6)
        m.fs.unit.membrane_charge_density.fix(-27)
        m.fs.unit.dielectric_constant_pore.fix(41.3)

        # Fix final permeate pressure to be ~atmospheric
        m.fs.unit.mixed_permeate[0].pressure.fix(101325)

        m.fs.unit.spacer_porosity.fix(0.85)
        m.fs.unit.channel_height.fix(5e-4)
        m.fs.unit.velocity[0, 0].fix(0.25)
        m.fs.unit.area.fix(50)

        return m

    @pytest.mark.unit
    def test_build(self, NF_frame):
        m = NF_frame

        # test ports and variables
        port_lst = ["inlet", "retentate", "permeate"]
        port_vars_lst = ["flow_mol_phase_comp", "pressure", "temperature"]
        for port_str in port_lst:
            assert hasattr(m.fs.unit, port_str)
            port = getattr(m.fs.unit, port_str)
            assert len(port.vars) == 3
            assert isinstance(port, Port)
            for var_str in port_vars_lst:
                assert hasattr(port, var_str)
                var = getattr(port, var_str)
                assert isinstance(var, Var)

        # test unit objects (including parameters, variables, and constraints)
        unit_objs_lst = []
        for obj_str in unit_objs_lst:
            assert hasattr(m.fs.unit, obj_str)

        # test state block objects
        assert isinstance(m.fs.unit.feed_side, ControlVolume0DBlock)
        cv_stateblock_lst = [
            "properties_in",
            "properties_out",
            "properties_interface",
        ]
        # feed side
        for sb_str in cv_stateblock_lst:
            sb = getattr(m.fs.unit.feed_side, sb_str)
            assert isinstance(sb, MCASStateBlock)
        # test objects added to control volume
        cv_objs_type_dict = {"eq_feed_interface_isothermal": Constraint}
        for obj_str, obj_type in cv_objs_type_dict.items():
            obj = getattr(m.fs.unit.feed_side, obj_str)
            assert isinstance(obj, obj_type)
        # permeate side
        assert isinstance(m.fs.unit.permeate_side, MCASStateBlock)
        assert isinstance(m.fs.unit.mixed_permeate, MCASStateBlock)
        # membrane
        assert isinstance(m.fs.unit.pore_entrance, MCASStateBlock)
        assert isinstance(m.fs.unit.pore_exit, MCASStateBlock)

        # test statistics
        assert number_variables(m) == 304
        assert number_total_constraints(m) == 272
        assert number_unused_variables(m) == 11

    @pytest.mark.unit
    def test_dof(self, NF_frame):
        m = NF_frame
        check_dof(m, fail_flag=True)

    @pytest.mark.unit
    def test_calculate_scaling(self, NF_frame):
        m = NF_frame

        m.fs.properties.set_default_scaling(
            "flow_mol_phase_comp", 1e2, index=("Liq", "Cl_-")
        )
        m.fs.properties.set_default_scaling(
            "flow_mol_phase_comp", 1e2, index=("Liq", "Na_+")
        )
        m.fs.properties.set_default_scaling(
            "flow_mol_phase_comp", 1e0, index=("Liq", "H2O")
        )

        calculate_scaling_factors(m)

        # check that all variables have scaling factors
        unscaled_var_list = list(unscaled_variables_generator(m.fs.unit))
        assert len(unscaled_var_list) == 0

    @pytest.mark.component
    def test_initialize(self, NF_frame):
        m = NF_frame
        initialization_tester(m, outlvl=idaeslog.DEBUG)

        badly_scaled_var_lst = list(
            badly_scaled_var_generator(m.fs.unit, small=1e-5, zero=1e-12)
        )
        for var, val in badly_scaled_var_lst:
            print(var.name, val)
        assert len(badly_scaled_var_lst) == 0

    @pytest.mark.component
    def test_solve(self, NF_frame):
        m = NF_frame
        results = solver.solve(m, tee=True)

        # Check for optimal solution
        assert_optimal_termination(results)

    @pytest.mark.component
    def test_conservation(self, NF_frame):
        m = NF_frame
        b = m.fs.unit
        comp_lst = m.fs.properties.solute_set

        flow_mass_inlet = sum(
            b.feed_side.properties_in[0].flow_mass_phase_comp["Liq", j]
            for j in comp_lst
        )
        flow_mass_retentate = sum(
            b.feed_side.properties_out[0].flow_mass_phase_comp["Liq", j]
            for j in comp_lst
        )
        flow_mass_permeate = sum(
            b.mixed_permeate[0].flow_mass_phase_comp["Liq", j] for j in comp_lst
        )

        assert (
            abs(value(flow_mass_inlet - flow_mass_retentate - flow_mass_permeate))
            <= 1e-6
        )

    @pytest.mark.component
    def test_solution(self, NF_frame):
        m = NF_frame

        mole_flux_dict = {
            "Na_+": 0.0010636,
            "Cl_-": 0.0010636,
            "H2O": 0.1313564,
        }
        for j, val in mole_flux_dict.items():
            assert pytest.approx(val, rel=1e-3) == value(
                m.fs.unit.flux_mol_phase_comp_avg[0, "Liq", j]
            )

        # TODO: subsequently focus on the segment below during the validation and refinement phase
        intrinsic_rejection_dict = {
            "Na_+": 0.096269,
            "Cl_-": 0.096269,
        }
        for j, val in intrinsic_rejection_dict.items():
            assert pytest.approx(val, rel=1e-3) == value(
                m.fs.unit.rejection_intrinsic_phase_comp[0, "Liq", j]
            )

    @pytest.mark.unit
    def test_report(self, NF_frame):
        NF_frame.fs.unit.report()


class TestNanoFiltration_with_CP_5ions_double_concentration:
    @pytest.fixture(scope="class")
    def NF_frame(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.properties = MCASParameterBlock(
            solute_list=["Ca_2+", "SO4_2-", "Mg_2+", "Na_+", "Cl_-"],
            diffusivity_data={
                ("Liq", "Ca_2+"): 9.2e-10,
                ("Liq", "SO4_2-"): 1.06e-09,
                ("Liq", "Mg_2+"): 7.06e-10,
                ("Liq", "Na_+"): 1.33e-09,
                ("Liq", "Cl_-"): 2.03e-09,
            },
            mw_data={
                "H2O": 0.018,
                "Ca_2+": 0.04,
                "Mg_2+": 0.024,
                "SO4_2-": 0.096,
                "Na_+": 0.023,
                "Cl_-": 0.035,
            },
            stokes_radius_data={
                "Ca_2+": 3.09e-10,
                "Mg_2+": 3.47e-10,
                "SO4_2-": 2.3e-10,
                "Cl_-": 1.21e-10,
                "Na_+": 1.84e-10,
            },
            charge={"Ca_2+": 2, "Mg_2+": 2, "SO4_2-": -2, "Na_+": 1, "Cl_-": -1},
            activity_coefficient_model=ActivityCoefficientModel.davies,
            density_calculation=DensityCalculation.constant,
        )

        m.fs.unit = NanofiltrationDSPMDE0D(property_package=m.fs.properties)
        mass_flow_in = 1 * pyunits.kg / pyunits.s
        feed_mass_frac = {
            "Ca_2+": 2 * 382e-6,
            "Mg_2+": 2 * 1394e-6,
            "SO4_2-": 2 * 2136e-6,
            "Cl_-": 2 * 20101.6e-6,
            "Na_+": 2 * 11122e-6,
        }

        # Fix mole flow rates of each ion and water
        for ion, x in feed_mass_frac.items():
            mol_comp_flow = (
                x
                * pyunits.kg
                / pyunits.kg
                * mass_flow_in
                / m.fs.unit.feed_side.properties_in[0].mw_comp[ion]
            )
            m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", ion].fix(mol_comp_flow)
        H2O_mass_frac = 1 - sum(x for x in feed_mass_frac.values())
        H2O_mol_comp_flow = (
            H2O_mass_frac
            * pyunits.kg
            / pyunits.kg
            * mass_flow_in
            / m.fs.unit.feed_side.properties_in[0].mw_comp["H2O"]
        )
        m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "H2O"].fix(H2O_mol_comp_flow)

        # Use assert electroneutrality method from property model to ensure the ion concentrations provided
        # obey electroneutrality condition
        m.fs.unit.feed_side.properties_in[0].assert_electroneutrality(
            defined_state=True,
            adjust_by_ion="Cl_-",
            get_property="mass_frac_phase_comp",
        )

        # Fix other inlet state variables
        m.fs.unit.inlet.temperature[0].fix(298.15)
        m.fs.unit.inlet.pressure[0].fix(4e5)

        # Fix the membrane variables that are usually fixed for the DSPM-DE model
        m.fs.unit.radius_pore.fix(0.5e-9)
        m.fs.unit.membrane_thickness_effective.fix(1.33e-6)
        m.fs.unit.membrane_charge_density.fix(-27)
        m.fs.unit.dielectric_constant_pore.fix(41.3)

        # Fix final permeate pressure to be ~atmospheric
        m.fs.unit.mixed_permeate[0].pressure.fix(101325)

        m.fs.unit.spacer_porosity.fix(0.85)
        m.fs.unit.channel_height.fix(5e-4)
        m.fs.unit.velocity[0, 0].fix(0.25)
        m.fs.unit.area.fix(50)
        # Fix additional variables for calculating mass transfer coefficient with spiral wound correlation
        m.fs.unit.spacer_mixing_efficiency.fix()
        m.fs.unit.spacer_mixing_length.fix()
        return m

    @pytest.mark.unit
    def test_build(self, NF_frame):
        m = NF_frame

        # test ports and variables
        port_lst = ["inlet", "retentate", "permeate"]
        port_vars_lst = ["flow_mol_phase_comp", "pressure", "temperature"]
        for port_str in port_lst:
            assert hasattr(m.fs.unit, port_str)
            port = getattr(m.fs.unit, port_str)
            assert len(port.vars) == 3
            assert isinstance(port, Port)
            for var_str in port_vars_lst:
                assert hasattr(port, var_str)
                var = getattr(port, var_str)
                assert isinstance(var, Var)

        # test unit objects (including parameters, variables, and constraints)
        unit_objs_lst = []
        for obj_str in unit_objs_lst:
            assert hasattr(m.fs.unit, obj_str)

        # test state block objects
        assert isinstance(m.fs.unit.feed_side, ControlVolume0DBlock)
        cv_stateblock_lst = [
            "properties_in",
            "properties_out",
            "properties_interface",
        ]
        # feed side
        for sb_str in cv_stateblock_lst:
            sb = getattr(m.fs.unit.feed_side, sb_str)
            assert isinstance(sb, MCASStateBlock)
        # test objects added to control volume
        cv_objs_type_dict = {"eq_feed_interface_isothermal": Constraint}
        for obj_str, obj_type in cv_objs_type_dict.items():
            obj = getattr(m.fs.unit.feed_side, obj_str)
            assert isinstance(obj, obj_type)
        # permeate side
        assert isinstance(m.fs.unit.permeate_side, MCASStateBlock)
        assert isinstance(m.fs.unit.mixed_permeate, MCASStateBlock)
        # membrane
        assert isinstance(m.fs.unit.pore_entrance, MCASStateBlock)
        assert isinstance(m.fs.unit.pore_exit, MCASStateBlock)

        # test statistics
        assert number_variables(m) == 566
        assert number_total_constraints(m) == 526
        assert number_unused_variables(m) == 11

    @pytest.mark.unit
    def test_dof(self, NF_frame):
        m = NF_frame
        check_dof(m, fail_flag=True)

    @pytest.mark.unit
    def test_calculate_scaling(self, NF_frame):
        m = NF_frame

        m.fs.properties.set_default_scaling(
            "flow_mol_phase_comp", 1e4, index=("Liq", "Ca_2+")
        )
        m.fs.properties.set_default_scaling(
            "flow_mol_phase_comp", 1e3, index=("Liq", "SO4_2-")
        )
        m.fs.properties.set_default_scaling(
            "flow_mol_phase_comp", 1e3, index=("Liq", "Mg_2+")
        )
        m.fs.properties.set_default_scaling(
            "flow_mol_phase_comp", 1e2, index=("Liq", "Cl_-")
        )
        m.fs.properties.set_default_scaling(
            "flow_mol_phase_comp", 1e2, index=("Liq", "Na_+")
        )
        m.fs.properties.set_default_scaling(
            "flow_mol_phase_comp", 1e0, index=("Liq", "H2O")
        )

        calculate_scaling_factors(m)

        # check that all variables have scaling factors
        unscaled_var_list = list(unscaled_variables_generator(m.fs.unit))
        assert len(unscaled_var_list) == 0

    @pytest.mark.component
    def test_initialize(self, NF_frame):
        m = NF_frame
        initialization_tester(m, outlvl=idaeslog.DEBUG)

        badly_scaled_var_lst = list(
            badly_scaled_var_generator(m.fs.unit, small=1e-5, zero=1e-12)
        )
        for var, val in badly_scaled_var_lst:
            print(var.name, val)
        assert len(badly_scaled_var_lst) == 0

    @pytest.mark.component
    def test_solve(self, NF_frame):
        m = NF_frame
        results = solver.solve(m, tee=True)

        # Check for optimal solution
        assert_optimal_termination(results)

    @pytest.mark.component
    def test_conservation(self, NF_frame):
        m = NF_frame
        b = m.fs.unit
        comp_lst = m.fs.properties.solute_set

        flow_mass_inlet = sum(
            b.feed_side.properties_in[0].flow_mass_phase_comp["Liq", j]
            for j in comp_lst
        )
        flow_mass_retentate = sum(
            b.feed_side.properties_out[0].flow_mass_phase_comp["Liq", j]
            for j in comp_lst
        )
        flow_mass_permeate = sum(
            b.mixed_permeate[0].flow_mass_phase_comp["Liq", j] for j in comp_lst
        )

        assert (
            abs(value(flow_mass_inlet - flow_mass_retentate - flow_mass_permeate))
            <= 1e-6
        )

    @pytest.mark.component
    def test_solution(self, NF_frame):
        m = NF_frame

        mole_flux_dict = {
            "Na_+": 0.0001269756,
            "Cl_-": 0.0001352,
            "Ca_2+": 1.4258536e-06,
            "SO4_2-": 2.9210354e-06,
            "Mg_2+": 5.5916186e-06,
            "H2O": 0.0069096,
        }
        for j, val in mole_flux_dict.items():
            assert pytest.approx(val, rel=1e-3) == value(
                m.fs.unit.flux_mol_phase_comp_avg[0, "Liq", j]
            )

        # TODO: subsequently focus on the segment below during the validation and refinement phase
        intrinsic_rejection_dict = {
            "Na_+": 0.00753,
            "Cl_-": 0.11317679,
            "Ca_2+": 0.441427,
            "SO4_2-": 0.510229,
            "Mg_2+": 0.64273,
        }
        for j, val in intrinsic_rejection_dict.items():
            assert pytest.approx(val, rel=1e-3) == value(
                m.fs.unit.rejection_intrinsic_phase_comp[0, "Liq", j]
            )

    @pytest.mark.unit
    def test_report(self, NF_frame):
        NF_frame.fs.unit.report()


@pytest.mark.component
def test_inverse_solve():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = MCASParameterBlock(
        solute_list=["Na_+", "Cl_-"],
        diffusivity_data={("Liq", "Na_+"): 1.33e-09, ("Liq", "Cl_-"): 2.03e-09},
        mw_data={"H2O": 0.018, "Na_+": 0.023, "Cl_-": 0.035},
        stokes_radius_data={"Cl_-": 1.21e-10, "Na_+": 1.84e-10},
        charge={"Na_+": 1, "Cl_-": -1},
        activity_coefficient_model=ActivityCoefficientModel.davies,
        density_calculation=DensityCalculation.constant,
    )

    m.fs.unit = NanofiltrationDSPMDE0D(property_package=m.fs.properties)

    m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "Na_+"].fix(0.429868)
    m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "Cl_-"].fix(0.429868)
    m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "H2O"].fix(47.356)

    # Fix other inlet state variables
    m.fs.unit.inlet.temperature[0].fix(298.15)
    m.fs.unit.inlet.pressure[0].fix(4e5)

    # Fix the membrane variables that are usually fixed for the DSPM-DE model
    m.fs.unit.radius_pore.fix(0.5e-9)
    m.fs.unit.membrane_thickness_effective.fix(1.33e-6)
    m.fs.unit.membrane_charge_density.fix(-27)
    m.fs.unit.dielectric_constant_pore.fix(41.3)

    # Fix final permeate pressure to be ~atmospheric
    m.fs.unit.mixed_permeate[0].pressure.fix(101325)

    m.fs.unit.spacer_porosity.fix(0.85)
    m.fs.unit.channel_height.fix(5e-4)
    m.fs.unit.velocity[0, 0].fix(0.25)
    m.fs.unit.area.fix(50)
    # Fix additional variables for calculating mass transfer coefficient with spiral wound correlation
    m.fs.unit.spacer_mixing_efficiency.fix()
    m.fs.unit.spacer_mixing_length.fix()

    check_dof(m, fail_flag=True)

    m.fs.properties.set_default_scaling(
        "flow_mol_phase_comp", 1e2, index=("Liq", "Cl_-")
    )
    m.fs.properties.set_default_scaling(
        "flow_mol_phase_comp", 1e2, index=("Liq", "Na_+")
    )
    m.fs.properties.set_default_scaling(
        "flow_mol_phase_comp", 1e0, index=("Liq", "H2O")
    )

    calculate_scaling_factors(m)

    # check that all variables have scaling factors
    unscaled_var_list = list(unscaled_variables_generator(m.fs.unit))
    assert len(unscaled_var_list) == 0

    initialization_tester(m, outlvl=idaeslog.DEBUG)

    badly_scaled_var_lst = list(
        badly_scaled_var_generator(m.fs.unit, small=1e-5, zero=1e-12)
    )
    for var, val in badly_scaled_var_lst:
        print(var.name, val)
    assert len(badly_scaled_var_lst) == 0

    m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "Na_+"].unfix()
    m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "Cl_-"].unfix()
    m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "H2O"].unfix()
    m.fs.unit.inlet.temperature[0].unfix()
    m.fs.unit.inlet.pressure[0].unfix()

    m.fs.unit.retentate.flow_mol_phase_comp[0, "Liq", "Na_+"].fix(0.429868)
    m.fs.unit.retentate.flow_mol_phase_comp[0, "Liq", "Cl_-"].fix(0.429868)
    m.fs.unit.retentate.flow_mol_phase_comp[0, "Liq", "H2O"].fix(47.356)
    m.fs.unit.retentate.temperature[0].fix(298.15)
    m.fs.unit.retentate.pressure[0].fix(4e5)

    results = solver.solve(m, tee=True)

    # Check for optimal solution
    assert_optimal_termination(results)

    b = m.fs.unit
    comp_lst = m.fs.properties.solute_set

    flow_mass_inlet = sum(
        b.feed_side.properties_in[0].flow_mass_phase_comp["Liq", j] for j in comp_lst
    )
    flow_mass_retentate = sum(
        b.feed_side.properties_out[0].flow_mass_phase_comp["Liq", j] for j in comp_lst
    )
    flow_mass_permeate = sum(
        b.mixed_permeate[0].flow_mass_phase_comp["Liq", j] for j in comp_lst
    )

    assert (
        abs(value(flow_mass_inlet - flow_mass_retentate - flow_mass_permeate)) <= 1e-6
    )

    mole_flux_dict = {
        "Na_+": 0.0010739,
        "Cl_-": 0.0010739,
        "H2O": 0.129305,
    }
    for j, val in mole_flux_dict.items():
        assert pytest.approx(val, rel=1e-3) == value(
            m.fs.unit.flux_mol_phase_comp_avg[0, "Liq", j]
        )

    intrinsic_rejection_dict = {
        "Na_+": 0.0964865,
        "Cl_-": 0.0964865,
    }
    for j, val in intrinsic_rejection_dict.items():
        assert pytest.approx(val, rel=1e-3) == value(
            m.fs.unit.rejection_intrinsic_phase_comp[0, "Liq", j]
        )
    flow_mol_in_dict = {"Na_+": 0.483564, "Cl_-": 0.483564, "H2O": 53.821}
    for j, val in flow_mol_in_dict.items():
        assert pytest.approx(val, rel=1e-3) == value(
            m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", j]
        )

    m.fs.unit.report()


@pytest.mark.component
def test_mass_transfer_coeff_fixed():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = MCASParameterBlock(
        solute_list=["Na_+", "Cl_-"],
        diffusivity_data={("Liq", "Na_+"): 1.33e-09, ("Liq", "Cl_-"): 2.03e-09},
        mw_data={"H2O": 0.018, "Na_+": 0.023, "Cl_-": 0.035},
        stokes_radius_data={"Cl_-": 1.21e-10, "Na_+": 1.84e-10},
        charge={"Na_+": 1, "Cl_-": -1},
        activity_coefficient_model=ActivityCoefficientModel.davies,
        density_calculation=DensityCalculation.constant,
    )

    m.fs.unit = NanofiltrationDSPMDE0D(
        property_package=m.fs.properties,
        mass_transfer_coefficient=MassTransferCoefficient.fixed,
    )

    m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "Na_+"].fix(0.429868)
    m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "Cl_-"].fix(0.429868)
    m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "H2O"].fix(47.356)

    # Fix other inlet state variables
    m.fs.unit.inlet.temperature[0].fix(298.15)
    m.fs.unit.inlet.pressure[0].fix(4e5)

    # Fix the membrane variables that are usually fixed for the DSPM-DE model
    m.fs.unit.radius_pore.fix(0.5e-9)
    m.fs.unit.membrane_thickness_effective.fix(1.33e-6)
    m.fs.unit.membrane_charge_density.fix(-27)
    m.fs.unit.dielectric_constant_pore.fix(41.3)

    # Fix final permeate pressure to be ~atmospheric
    m.fs.unit.mixed_permeate[0].pressure.fix(101325)

    m.fs.unit.spacer_porosity.fix(0.85)
    m.fs.unit.channel_height.fix(5e-4)
    m.fs.unit.velocity[0, 0].fix(0.25)
    m.fs.unit.area.fix(50)

    # Fix mass transfer coefficient for each ion
    m.fs.unit.Kf_comp.fix(1e-5)

    check_dof(m, fail_flag=True)

    m.fs.properties.set_default_scaling(
        "flow_mol_phase_comp", 1e2, index=("Liq", "Cl_-")
    )
    m.fs.properties.set_default_scaling(
        "flow_mol_phase_comp", 1e2, index=("Liq", "Na_+")
    )
    m.fs.properties.set_default_scaling(
        "flow_mol_phase_comp", 1e0, index=("Liq", "H2O")
    )

    calculate_scaling_factors(m)

    # check that all variables have scaling factors
    unscaled_var_list = list(unscaled_variables_generator(m.fs.unit))
    assert len(unscaled_var_list) == 0

    initialization_tester(m, outlvl=idaeslog.DEBUG)

    badly_scaled_var_lst = list(
        badly_scaled_var_generator(m.fs.unit, small=1e-5, zero=1e-12)
    )
    for var, val in badly_scaled_var_lst:
        print(var.name, val)
    assert len(badly_scaled_var_lst) == 0

    m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "Na_+"].unfix()
    m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "Cl_-"].unfix()
    m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "H2O"].unfix()
    m.fs.unit.inlet.temperature[0].unfix()
    m.fs.unit.inlet.pressure[0].unfix()

    m.fs.unit.retentate.flow_mol_phase_comp[0, "Liq", "Na_+"].fix(0.429868)
    m.fs.unit.retentate.flow_mol_phase_comp[0, "Liq", "Cl_-"].fix(0.429868)
    m.fs.unit.retentate.flow_mol_phase_comp[0, "Liq", "H2O"].fix(47.356)
    m.fs.unit.retentate.temperature[0].fix(298.15)
    m.fs.unit.retentate.pressure[0].fix(4e5)

    results = solver.solve(m, tee=True)

    # Check for optimal solution
    assert_optimal_termination(results)

    b = m.fs.unit
    comp_lst = m.fs.properties.solute_set

    flow_mass_inlet = sum(
        b.feed_side.properties_in[0].flow_mass_phase_comp["Liq", j] for j in comp_lst
    )
    flow_mass_retentate = sum(
        b.feed_side.properties_out[0].flow_mass_phase_comp["Liq", j] for j in comp_lst
    )
    flow_mass_permeate = sum(
        b.mixed_permeate[0].flow_mass_phase_comp["Liq", j] for j in comp_lst
    )

    assert (
        abs(value(flow_mass_inlet - flow_mass_retentate - flow_mass_permeate)) <= 1e-6
    )

    mole_flux_dict = {
        "Na_+": 0.00107097,
        "Cl_-": 0.00107097,
        "H2O": 0.12989686,
    }
    for j, val in mole_flux_dict.items():
        assert pytest.approx(val, rel=1e-3) == value(
            m.fs.unit.flux_mol_phase_comp_avg[0, "Liq", j]
        )

    intrinsic_rejection_dict = {
        "Na_+": 0.09700,
        "Cl_-": 0.09700,
    }
    for j, val in intrinsic_rejection_dict.items():
        assert pytest.approx(val, rel=1e-3) == value(
            m.fs.unit.rejection_intrinsic_phase_comp[0, "Liq", j]
        )
    flow_mol_in_dict = {"Na_+": 0.483564, "Cl_-": 0.483564, "H2O": 53.821}
    for j, val in flow_mol_in_dict.items():
        assert pytest.approx(val, rel=1e-3) == value(
            m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", j]
        )


@pytest.mark.unit
def test_mass_transfer_CP_config_errors():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = MCASParameterBlock(
        solute_list=["Na_+", "Cl_-"],
        diffusivity_data={("Liq", "Na_+"): 1.33e-09, ("Liq", "Cl_-"): 2.03e-09},
        mw_data={"H2O": 0.018, "Na_+": 0.023, "Cl_-": 0.035},
        stokes_radius_data={"Cl_-": 1.21e-10, "Na_+": 1.84e-10},
        charge={"Na_+": 1, "Cl_-": -1},
        activity_coefficient_model=ActivityCoefficientModel.davies,
        density_calculation=DensityCalculation.constant,
    )

    with pytest.raises(
        ConfigurationError,
        match="\nConflict between configuration options:\n"
        "'mass_transfer_coefficient' cannot be set to MassTransferCoefficient.fixed "
        "while 'concentration_polarization_type' is set to ConcentrationPolarizationType.none.\n\n"
        "'mass_transfer_coefficient' must be set to MassTransferCoefficient.none\nor "
        "'concentration_polarization_type' must be set to ConcentrationPolarizationType.calculated",
    ):
        m.fs.unit = NanofiltrationDSPMDE0D(
            property_package=m.fs.properties,
            mass_transfer_coefficient=MassTransferCoefficient.fixed,
            concentration_polarization_type=ConcentrationPolarizationType.none,
        )

    with pytest.raises(
        ConfigurationError,
        match="\nConflict between configuration options:\n"
        "'mass_transfer_coefficient' cannot be set to MassTransferCoefficient.none "
        "while 'concentration_polarization_type' is set to ConcentrationPolarizationType.calculated.\n\n"
        "'mass_transfer_coefficient' must be set to MassTransferCoefficient.none\nor "
        "'concentration_polarization_type' must be set to ConcentrationPolarizationType.calculated",
    ):
        m.fs.unit = NanofiltrationDSPMDE0D(
            property_package=m.fs.properties,
            mass_transfer_coefficient=MassTransferCoefficient.none,
            concentration_polarization_type=ConcentrationPolarizationType.calculated,
        )

    with pytest.raises(
        ConfigurationError,
        match="\nConflict between configuration options:\n"
        "'mass_transfer_coefficient' cannot be set to MassTransferCoefficient.spiral_wound "
        "while 'concentration_polarization_type' is set to ConcentrationPolarizationType.none.\n\n"
        "'mass_transfer_coefficient' must be set to MassTransferCoefficient.none\nor "
        "'concentration_polarization_type' must be set to ConcentrationPolarizationType.calculated",
    ):
        m.fs.unit = NanofiltrationDSPMDE0D(
            property_package=m.fs.properties,
            mass_transfer_coefficient=MassTransferCoefficient.spiral_wound,
            concentration_polarization_type=ConcentrationPolarizationType.none,
        )


@pytest.mark.component
def test_pressure_recovery_step_2_ions():
    "Test optimal termination across a range of pressures and recovery rates for 2 ion system"
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = MCASParameterBlock(
        solute_list=["Na_+", "Cl_-"],
        diffusivity_data={("Liq", "Na_+"): 1.33e-09, ("Liq", "Cl_-"): 2.03e-09},
        mw_data={"H2O": 0.018, "Na_+": 0.023, "Cl_-": 0.035},
        stokes_radius_data={"Cl_-": 1.21e-10, "Na_+": 1.84e-10},
        charge={"Na_+": 1, "Cl_-": -1},
        activity_coefficient_model=ActivityCoefficientModel.davies,
        density_calculation=DensityCalculation.constant,
    )

    m.fs.unit = NanofiltrationDSPMDE0D(property_package=m.fs.properties)

    m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "Na_+"].fix(0.429868)
    m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "Cl_-"].fix(0.429868)
    m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "H2O"].fix(47.356)

    # Fix other inlet state variables
    m.fs.unit.inlet.temperature[0].fix(298.15)
    m.fs.unit.inlet.pressure[0].fix(4e5)

    # Fix the membrane variables that are usually fixed for the DSPM-DE model
    m.fs.unit.radius_pore.fix(0.5e-9)
    m.fs.unit.membrane_thickness_effective.fix(1.33e-6)
    m.fs.unit.membrane_charge_density.fix(-27)
    m.fs.unit.dielectric_constant_pore.fix(41.3)

    # Fix final permeate pressure to be ~atmospheric
    m.fs.unit.mixed_permeate[0].pressure.fix(101325)

    m.fs.unit.spacer_porosity.fix(0.85)
    m.fs.unit.channel_height.fix(5e-4)
    m.fs.unit.velocity[0, 0].fix(0.25)
    m.fs.unit.area.fix(50)
    # Fix additional variables for calculating mass transfer coefficient with spiral wound correlation
    m.fs.unit.spacer_mixing_efficiency.fix()
    m.fs.unit.spacer_mixing_length.fix()

    check_dof(m, fail_flag=True)

    m.fs.properties.set_default_scaling(
        "flow_mol_phase_comp", 1e2, index=("Liq", "Cl_-")
    )
    m.fs.properties.set_default_scaling(
        "flow_mol_phase_comp", 1e2, index=("Liq", "Na_+")
    )
    m.fs.properties.set_default_scaling(
        "flow_mol_phase_comp", 1e0, index=("Liq", "H2O")
    )

    calculate_scaling_factors(m)

    # check that all variables have scaling factors
    unscaled_var_list = list(unscaled_variables_generator(m.fs.unit))
    assert len(unscaled_var_list) == 0

    # Expect only flux_mol_phase_comp to be poorly scaled, as we have not
    # calculated correct values just yet.
    for var in list(badly_scaled_var_generator(m.fs.unit)):
        assert "flux_mol_phase_comp" in var[0].name

    initialization_tester(m)

    results = solver.solve(m)

    # Check for optimal solution
    assert_optimal_termination(results)

    pressure_steps = np.linspace(1.8e5, 20e5, 10)

    for p in pressure_steps:
        m.fs.unit.inlet.pressure[0].fix(p)
        results = solver.solve(m)
        assert_optimal_termination(results)

    m.fs.unit.inlet.pressure[0].fix(10e5)
    m.fs.unit.area.unfix()

    for r in np.linspace(0.05, 0.97, 10):
        m.fs.unit.recovery_vol_phase.fix(r)
        print(r)
        res = solver.solve(m, tee=True)
        assert_optimal_termination(res)


def calc_scale(value):
    return -1 * log(value, 10)


@pytest.mark.component
def test_pressure_recovery_step_5_ions():
    "Test optimal termination across a range of pressures and recovery rates for 5 ion system"
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    property_kwds = {
        "solute_list": [
            "Ca_2+",
            "SO4_2-",
            "HCO3_-",
            "Na_+",
            "Cl_-",
        ],
        "diffusivity_data": {
            ("Liq", "Ca_2+"): 9.2e-10,
            ("Liq", "SO4_2-"): 1.06e-9,
            ("Liq", "HCO3_-"): 1.19e-9,
            ("Liq", "Na_+"): 1.33e-9,
            ("Liq", "Cl_-"): 2.03e-9,
        },
        "mw_data": {
            "H2O": 18e-3,
            "Ca_2+": 40e-3,
            "HCO3_-": 61.0168e-3,
            "SO4_2-": 96e-3,
            "Na_+": 23e-3,
            "Cl_-": 35e-3,
        },
        "stokes_radius_data": {
            "Ca_2+": 0.309e-9,
            "HCO3_-": 2.06e-10,
            "SO4_2-": 0.230e-9,
            "Cl_-": 0.121e-9,
            "Na_+": 0.184e-9,
        },
        "charge": {
            "Ca_2+": 2,
            "HCO3_-": -1,
            "SO4_2-": -2,
            "Na_+": 1,
            "Cl_-": -1,
        },
        "activity_coefficient_model": ActivityCoefficientModel.ideal,
        "density_calculation": DensityCalculation.constant,
    }

    m.fs.properties = MCASParameterBlock(**property_kwds)

    m.fs.feed = Feed(property_package=m.fs.properties)

    m.fs.nfUnit = NanofiltrationDSPMDE0D(property_package=m.fs.properties)

    m.fs.feed_to_nf = Arc(source=m.fs.feed.outlet, destination=m.fs.nfUnit.inlet)
    TransformationFactory("network.expand_arcs").apply_to(m)

    feed_mass_frac = {
        "Ca_2+": 4.0034374454637006e-04,
        "HCO3_-": 0.00022696833343821863,
        "SO4_2-": 0.00020497140244420624,
        "Cl_-": 0.0004559124032433401,
        "Na_+": 0.00043333830389924205,
    }

    mass_flow_in = 1 * pyunits.kg / pyunits.s
    for ion, x in feed_mass_frac.items():
        mol_comp_flow = (
            x
            * pyunits.kg
            / pyunits.kg
            * mass_flow_in
            / m.fs.feed.properties[0].mw_comp[ion]
        )
        m.fs.feed.properties[0].flow_mol_phase_comp["Liq", ion].fix(mol_comp_flow)
    H2O_mass_frac = 1 - sum(x for x in feed_mass_frac.values())
    H2O_mol_comp_flow = (
        H2O_mass_frac
        * pyunits.kg
        / pyunits.kg
        * mass_flow_in
        / m.fs.feed.properties[0].mw_comp["H2O"]
    )
    m.fs.feed.properties[0].flow_mol_phase_comp["Liq", "H2O"].fix(H2O_mol_comp_flow)

    for index in m.fs.feed.properties[0].flow_mol_phase_comp:
        scale = calc_scale(m.fs.feed.properties[0].flow_mol_phase_comp[index].value)
        m.fs.properties.set_default_scaling(
            "flow_mol_phase_comp", 10 ** (scale), index=index
        )
    m.fs.feed.properties[0].assert_electroneutrality(
        defined_state=True,
        adjust_by_ion="Cl_-",
        get_property="flow_mol_phase_comp",
    )

    m.fs.feed.properties[0].temperature.fix(298.15)

    calculate_scaling_factors(m.fs)

    m.fs.feed.initialize(optarg=solver.options)
    m.fs.feed.properties[0].pressure.fix(1.5 * 1e5)
    propagate_state(m.fs.feed_to_nf)
    m.fs.nfUnit.recovery_vol_phase.fix(0.1)

    m.fs.nfUnit.spacer_porosity.fix(0.85)
    m.fs.nfUnit.channel_height.fix(1e-3)
    m.fs.nfUnit.velocity[0, 0].fix(0.25)
    m.fs.nfUnit.spacer_mixing_efficiency.fix()
    m.fs.nfUnit.spacer_mixing_length.fix()

    m.fs.nfUnit.radius_pore.fix(0.5e-9)
    m.fs.nfUnit.membrane_thickness_effective.fix(8.6e-07)
    m.fs.nfUnit.membrane_charge_density.fix(-680)
    m.fs.nfUnit.dielectric_constant_pore.fix(41.3)
    m.fs.nfUnit.mixed_permeate[0].pressure.fix(101325)

    check_dof(m, fail_flag=True)

    m.fs.nfUnit.initialize(
        optarg=solver.options,
        automate_rescale=False,
    )

    res = solver.solve(m, tee=True)
    assert_optimal_termination(res)

    for k in np.linspace(2, 20, 10):
        m.fs.feed.properties[0].pressure.fix(k * 1e5)
        res = solver.solve(m, tee=True)
        assert_optimal_termination(res)

    m.fs.feed.properties[0].pressure.fix(10e5)

    for r in np.linspace(0.05, 0.97, 10):
        m.fs.nfUnit.recovery_vol_phase.fix(r)
        res = solver.solve(m, tee=True)
        assert_optimal_termination(res)



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pytest
from pyomo.environ import value, assert_optimal_termination
from watertap.examples.flowsheets.case_studies.wastewater_resource_recovery.suboxic_activated_sludge_process.suboxic_activated_sludge_process import (
    main,
)

# -----------------------------------------------------------------------------
@pytest.mark.component
def test_suboxicASM():
    m, results = main()

    assert_optimal_termination(results)

    assert value(m.fs.feed.properties[0].flow_mass_comp["H2O"]) == pytest.approx(
        355.3038, rel=1e-3
    )
    assert value(m.fs.feed.properties[0].flow_mass_comp["bod"]) == pytest.approx(
        0.1116, rel=1e-3
    )
    assert value(m.fs.feed.properties[0].flow_mass_comp["tss"]) == pytest.approx(
        0.1195, rel=1e-3
    )
    assert value(m.fs.feed.properties[0].flow_mass_comp["tkn"]) == pytest.approx(
        0.01849, rel=1e-3
    )
    assert value(m.fs.feed.properties[0].flow_mass_comp["phosphorus"]) == pytest.approx(
        0.002133, rel=1e-3
    )

    assert value(
        m.fs.suboxicASM.properties_treated[0].flow_mass_comp["H2O"]
    ) == pytest.approx(355.3038, rel=1e-3)
    assert value(
        m.fs.suboxicASM.properties_treated[0].flow_mass_comp["bod"]
    ) == pytest.approx(0.003550, rel=1e-3)
    assert value(
        m.fs.suboxicASM.properties_treated[0].flow_mass_comp["tss"]
    ) == pytest.approx(0.001780, rel=1e-3)
    assert value(
        m.fs.suboxicASM.properties_treated[0].flow_mass_comp["tkn"]
    ) == pytest.approx(0.002134, rel=1e-3)
    assert value(
        m.fs.suboxicASM.properties_treated[0].flow_mass_comp["phosphorus"]
    ) == pytest.approx(0.0003556, rel=1e-3)

    assert value(m.fs.costing.LCOW) == pytest.approx(0.660649, rel=1e-3)  # in $/m**3
    assert value(m.fs.costing.electricity_intensity) == pytest.approx(0.5003, rel=1e-3)



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Example of activated sludge process model based on BSM1 study using ASM1 reactions.
This process contains 2 anoxic tanks followed by 3 aerobic tanks and a clarifier
with recycle streams.

Unit operations are modeled as follows:

    * Anoxic reactors as standard CSTRs (CSTR)
    * Aerobic reactors as CSTRs with additional injection terms (CSTR_Injection)
    * Clarifier as a Separator with split fractions by components

Note that a pressure changer is required in the recycle stream to ensure the
pressure inside the recycle loop is bounded. As the inlet Mixer uses a pressure
minimization constraint and there is no pressure drop in the reactors, if pressure
is not specified at some point within the recycle loop then it becomes unbounded.

Based on example from:

[1] J. Alex, L. Benedetti, J. Copp, K.V. Gernaey, U. Jeppsson, I. Nopens, M.N. Pons,
J.P. Steyer and P. Vanrolleghem, "Benchmark Simulation Model no. 1 (BSM1)", 2018
"""

# Some more information about this module
__author__ = "Andrew Lee, Xinhong Liu"

import pyomo.environ as pyo
from pyomo.network import Arc, SequentialDecomposition

from idaes.core import FlowsheetBlock
from idaes.models.unit_models import (
    CSTR,
    Feed,
    Mixer,
    Separator,
    PressureChanger,
    Product,
)
from idaes.models.unit_models.separator import SplittingType
from idaes.core.solvers import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
import idaes.logger as idaeslog
import idaes.core.util.scaling as iscale
from idaes.core.util.tables import (
    create_stream_table_dataframe,
    stream_table_dataframe_to_string,
)

from watertap.unit_models.cstr_injection import CSTR_Injection
from watertap.property_models.activated_sludge.asm1_properties import ASM1ParameterBlock
from watertap.property_models.activated_sludge.asm1_reactions import (
    ASM1ReactionParameterBlock,
)

from watertap.core.util.initialization import check_solve

# Set up logger
_log = idaeslog.getLogger(__name__)


def build_flowsheet():
    m = pyo.ConcreteModel()

    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.props = ASM1ParameterBlock()
    m.fs.rxn_props = ASM1ReactionParameterBlock(property_package=m.fs.props)
    # Feed water stream
    m.fs.FeedWater = Feed(property_package=m.fs.props)
    # Mixer for feed water and recycled sludge
    m.fs.MX1 = Mixer(property_package=m.fs.props, inlet_list=["feed_water", "recycle"])
    # First reactor (anoxic) - standard CSTR
    m.fs.R1 = CSTR(property_package=m.fs.props, reaction_package=m.fs.rxn_props)
    # Second reactor (anoxic) - standard CSTR
    m.fs.R2 = CSTR(property_package=m.fs.props, reaction_package=m.fs.rxn_props)
    # Third reactor (aerobic) - CSTR with injection
    m.fs.R3 = CSTR_Injection(
        property_package=m.fs.props, reaction_package=m.fs.rxn_props
    )
    # Fourth reactor (aerobic) - CSTR with injection
    m.fs.R4 = CSTR_Injection(
        property_package=m.fs.props, reaction_package=m.fs.rxn_props
    )
    # Fifth reactor (aerobic) - CSTR with injection
    m.fs.R5 = CSTR_Injection(
        property_package=m.fs.props, reaction_package=m.fs.rxn_props
    )
    m.fs.SP5 = Separator(
        property_package=m.fs.props, outlet_list=["underflow", "overflow"]
    )
    # Clarifier
    # TODO: Replace with more detailed model when available
    m.fs.CL1 = Separator(
        property_package=m.fs.props,
        outlet_list=["underflow", "effluent"],
        split_basis=SplittingType.componentFlow,
    )
    # Sludge purge splitter
    m.fs.SP6 = Separator(
        property_package=m.fs.props,
        outlet_list=["recycle", "waste"],
        split_basis=SplittingType.totalFlow,
    )
    # Mixing sludge recycle and R5 underflow
    m.fs.MX6 = Mixer(property_package=m.fs.props, inlet_list=["clarifier", "reactor"])
    # Product Blocks
    m.fs.Treated = Product(property_package=m.fs.props)
    m.fs.Sludge = Product(property_package=m.fs.props)
    # Recycle pressure changer - use a simple isothermal unit for now
    m.fs.P1 = PressureChanger(property_package=m.fs.props)

    # Link units
    m.fs.stream1 = Arc(source=m.fs.FeedWater.outlet, destination=m.fs.MX1.feed_water)
    m.fs.stream2 = Arc(source=m.fs.MX1.outlet, destination=m.fs.R1.inlet)
    m.fs.stream3 = Arc(source=m.fs.R1.outlet, destination=m.fs.R2.inlet)
    m.fs.stream4 = Arc(source=m.fs.R2.outlet, destination=m.fs.R3.inlet)
    m.fs.stream5 = Arc(source=m.fs.R3.outlet, destination=m.fs.R4.inlet)
    m.fs.stream6 = Arc(source=m.fs.R4.outlet, destination=m.fs.R5.inlet)
    m.fs.stream7 = Arc(source=m.fs.R5.outlet, destination=m.fs.SP5.inlet)
    m.fs.stream8 = Arc(source=m.fs.SP5.overflow, destination=m.fs.CL1.inlet)
    m.fs.stream9 = Arc(source=m.fs.SP5.underflow, destination=m.fs.MX6.reactor)
    m.fs.stream10 = Arc(source=m.fs.CL1.effluent, destination=m.fs.Treated.inlet)
    m.fs.stream11 = Arc(source=m.fs.CL1.underflow, destination=m.fs.SP6.inlet)
    m.fs.stream102 = Arc(source=m.fs.SP6.waste, destination=m.fs.Sludge.inlet)
    m.fs.stream13 = Arc(source=m.fs.SP6.recycle, destination=m.fs.MX6.clarifier)
    m.fs.stream14 = Arc(source=m.fs.MX6.outlet, destination=m.fs.P1.inlet)
    m.fs.stream15 = Arc(source=m.fs.P1.outlet, destination=m.fs.MX1.recycle)
    pyo.TransformationFactory("network.expand_arcs").apply_to(m)

    # Oxygen concentration in reactors 3 and 4 is governed by mass transfer
    # Add additional parameter and constraints
    m.fs.R3.KLa = pyo.Var(
        initialize=7.6,
        units=pyo.units.hour**-1,
        doc="Lumped mass transfer coefficient for oxygen",
    )
    m.fs.R4.KLa = pyo.Var(
        initialize=5.7,
        units=pyo.units.hour**-1,
        doc="Lumped mass transfer coefficient for oxygen",
    )
    m.fs.S_O_eq = pyo.Param(
        default=8e-3,
        units=pyo.units.kg / pyo.units.m**3,
        mutable=True,
        doc="Dissolved oxygen concentration at equilibrium",
    )

    @m.fs.R3.Constraint(m.fs.time, doc="Mass transfer constraint for R3")
    def mass_transfer_R3(self, t):
        return pyo.units.convert(
            m.fs.R3.injection[t, "Liq", "S_O"], to_units=pyo.units.kg / pyo.units.hour
        ) == (
            m.fs.R3.KLa
            * m.fs.R3.volume[t]
            * (m.fs.S_O_eq - m.fs.R3.outlet.conc_mass_comp[t, "S_O"])
        )

    @m.fs.R4.Constraint(m.fs.time, doc="Mass transfer constraint for R4")
    def mass_transfer_R4(self, t):
        return pyo.units.convert(
            m.fs.R4.injection[t, "Liq", "S_O"], to_units=pyo.units.kg / pyo.units.hour
        ) == (
            m.fs.R4.KLa
            * m.fs.R4.volume[t]
            * (m.fs.S_O_eq - m.fs.R4.outlet.conc_mass_comp[t, "S_O"])
        )

    # Feed Water Conditions
    m.fs.FeedWater.flow_vol.fix(18446 * pyo.units.m**3 / pyo.units.day)
    m.fs.FeedWater.temperature.fix(298.15 * pyo.units.K)
    m.fs.FeedWater.pressure.fix(1 * pyo.units.atm)
    m.fs.FeedWater.conc_mass_comp[0, "S_I"].fix(30 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "S_S"].fix(69.5 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "X_I"].fix(51.2 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "X_S"].fix(202.32 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "X_BH"].fix(28.17 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "X_BA"].fix(0 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "X_P"].fix(0 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "S_O"].fix(0 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "S_NO"].fix(0 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "S_NH"].fix(31.56 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "S_ND"].fix(6.95 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "X_ND"].fix(10.59 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.alkalinity.fix(7 * pyo.units.mol / pyo.units.m**3)

    # Reactor sizing
    m.fs.R1.volume.fix(1000 * pyo.units.m**3)
    m.fs.R2.volume.fix(1000 * pyo.units.m**3)
    m.fs.R3.volume.fix(1333 * pyo.units.m**3)
    m.fs.R4.volume.fix(1333 * pyo.units.m**3)
    m.fs.R5.volume.fix(1333 * pyo.units.m**3)

    # Injection rates to Reactions 3, 4 and 5
    for j in m.fs.props.component_list:
        if j != "S_O":
            # All components except S_O have no injection
            m.fs.R3.injection[:, :, j].fix(0)
            m.fs.R4.injection[:, :, j].fix(0)
            m.fs.R5.injection[:, :, j].fix(0)
    # Then set injections rates for O2
    m.fs.R3.outlet.conc_mass_comp[:, "S_O"].fix(1.72e-3)
    m.fs.R4.outlet.conc_mass_comp[:, "S_O"].fix(2.43e-3)
    m.fs.R5.outlet.conc_mass_comp[:, "S_O"].fix(4.49e-4)

    # Set fraction of outflow from reactor 5 that goes to recycle
    m.fs.SP5.split_fraction[:, "underflow"].fix(0.6)

    # Clarifier
    # TODO: Update once more detailed model available
    m.fs.CL1.split_fraction[0, "effluent", "H2O"].fix(0.48956)
    m.fs.CL1.split_fraction[0, "effluent", "S_I"].fix(0.48956)
    m.fs.CL1.split_fraction[0, "effluent", "S_S"].fix(0.48956)
    m.fs.CL1.split_fraction[0, "effluent", "X_I"].fix(0.00187)
    m.fs.CL1.split_fraction[0, "effluent", "X_S"].fix(0.00187)
    m.fs.CL1.split_fraction[0, "effluent", "X_BH"].fix(0.00187)
    m.fs.CL1.split_fraction[0, "effluent", "X_BA"].fix(0.00187)
    m.fs.CL1.split_fraction[0, "effluent", "X_P"].fix(0.00187)
    m.fs.CL1.split_fraction[0, "effluent", "S_O"].fix(0.48956)
    m.fs.CL1.split_fraction[0, "effluent", "S_NO"].fix(0.48956)
    m.fs.CL1.split_fraction[0, "effluent", "S_NH"].fix(0.48956)
    m.fs.CL1.split_fraction[0, "effluent", "S_ND"].fix(0.48956)
    m.fs.CL1.split_fraction[0, "effluent", "X_ND"].fix(0.00187)
    m.fs.CL1.split_fraction[0, "effluent", "S_ALK"].fix(0.48956)

    # Sludge purge separator
    m.fs.SP6.split_fraction[:, "recycle"].fix(0.97955)

    # Outlet pressure from recycle pump
    m.fs.P1.outlet.pressure.fix(101325)

    # Check degrees of freedom
    print(degrees_of_freedom(m))
    assert degrees_of_freedom(m) == 0

    # Apply scaling
    iscale.calculate_scaling_factors(m.fs)

    # Initialize flowsheet
    # Apply sequential decomposition - 1 iteration should suffice
    seq = SequentialDecomposition()
    seq.options.select_tear_method = "heuristic"
    seq.options.tear_method = "Wegstein"
    seq.options.iterLim = 1

    G = seq.create_graph(m)

    # # Uncomment this code to see tear set and initialization order
    # heuristic_tear_set = seq.tear_set_arcs(G, method="heuristic")
    # order = seq.calculation_order(G)
    # for o in heuristic_tear_set:
    #     print(o.name)
    # for o in order:
    #     print(o[0].name)

    # Initial guesses for flow into first reactor
    tear_guesses = {
        "flow_vol": {0: 1.0675},
        "conc_mass_comp": {
            (0, "S_I"): 0.03,
            (0, "S_S"): 0.0146,
            (0, "X_I"): 1.149,
            (0, "X_S"): 0.0893,
            (0, "X_BH"): 2.542,
            (0, "X_BA"): 0.149,
            (0, "X_P"): 0.448,
            (0, "S_O"): 3.93e-4,
            (0, "S_NO"): 8.32e-3,
            (0, "S_NH"): 7.7e-3,
            (0, "S_ND"): 1.94e-3,
            (0, "X_ND"): 5.62e-3,
        },
        "alkalinity": {0: 5e-3},
        "temperature": {0: 298.15},
        "pressure": {0: 101325},
    }

    # Pass the tear_guess to the SD tool
    seq.set_guesses_for(m.fs.R1.inlet, tear_guesses)

    def function(unit):
        unit.initialize(outlvl=idaeslog.INFO)

    seq.run(m, function)

    # Solve overall flowsheet to close recycle loop
    solver = get_solver()
    results = solver.solve(m)
    check_solve(results, checkpoint="closing recycle", logger=_log, fail_flag=True)

    # Switch to fixed KLa in R3 and R4 (S_O concentration is controlled in R5)
    m.fs.R3.KLa.fix(10)
    m.fs.R4.KLa.fix(10)
    m.fs.R3.outlet.conc_mass_comp[:, "S_O"].unfix()
    m.fs.R4.outlet.conc_mass_comp[:, "S_O"].unfix()

    # Resolve with controls in place
    results = solver.solve(m, tee=True)
    check_solve(
        results,
        checkpoint="re-solve with controls in place",
        logger=_log,
        fail_flag=True,
    )

    return m, results


if __name__ == "__main__":
    # This method builds and runs a steady state activated sludge
    # flowsheet.
    m, results = build_flowsheet()

    stream_table = create_stream_table_dataframe(
        {
            "Feed": m.fs.FeedWater.outlet,
            "R1": m.fs.R1.outlet,
            "R2": m.fs.R2.outlet,
            "R3": m.fs.R3.outlet,
            "R4": m.fs.R4.outlet,
            "R5": m.fs.R5.outlet,
        },
        time_point=0,
    )
    print(stream_table_dataframe_to_string(stream_table))



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
import contextlib
from dataclasses import dataclass, field, asdict
import logging
import os
from pathlib import Path
import shutil
from typing import List, Optional, Tuple, Union, Any

import pytest
from click import Command
from click.testing import CliRunner, Result
import pymongo

from watertap.edb import commands, ElectrolyteDB


@pytest.fixture(scope="module")
def runner():
    return CliRunner()


# TODO move this to edb.commands

LogLevel = type(logging.INFO)
ExitCode = commands.ExitCode


@dataclass
class Expected:
    exit_code: ExitCode = ExitCode.OK
    file_name: Optional[str] = None
    file_path: Optional[Path] = None
    log_level: Optional[LogLevel] = None

    @property
    def success(self) -> bool:
        return self.exit_code == ExitCode.OK

    def __post_init__(self):
        if self.file_name is not None and self.file_path is None:
            self.file_path = Path(self.file_name)

    def __repr__(self):
        kv_repr = [f"{k}={v!r}" for k, v in asdict(self).items() if v is not None]
        return f"{type(self).__name__}" f"({', '.join(kv_repr)})"

    def __eq__(self, other: Union["Expected", Result]):
        if isinstance(other, Result):
            return self.exit_code == other.exit_code
        return type(self) == type(other) and asdict(self) == asdict(other)


Outcome = Tuple[Result, Expected]


@dataclass
class Invocation:
    args: List[str]
    expected: Optional[Expected] = field(default_factory=Expected)
    command: Optional[Command] = commands.command_base

    def run(self, runner: CliRunner) -> Result:
        runner = runner or CliRunner()
        return runner.invoke(self.command, self.args)
        self.result = result
        return result


def _display_param(obj: Any) -> str:
    if isinstance(obj, Invocation):
        args = " ".join(obj.args)
        return f"{args!r}, {obj.expected}"
    if isinstance(obj, tuple):
        return _display_param(Invocation(*obj))
    return str(obj)


class _INVALID:
    url: str = "__INVALID__"
    file_name: str = "__INVALID__"
    data_type: str = "__INVALID__"


EDBClient = pymongo.MongoClient


@dataclass
class EDBClientFactory:
    instance: EDBClient = None

    def __call__(self, url=None, **kwargs) -> EDBClient:
        if url == _INVALID.url:
            raise pymongo.errors.ConnectionFailure(f"Invalid url: {url}")
        return self.instance


@pytest.fixture(scope="function")
def mock_edb(monkeypatch) -> EDBClientFactory:
    mongomock = pytest.importorskip(
        "mongomock", reason="mongomock (EDB optional dependency) not available"
    )

    # NOTE since mongomock clients store data in memory,
    # the same MongoClient instance must be kept and used for the lifetime of the fixture
    # to be able to access the EDB data for e.g. assertions
    client_factory = EDBClientFactory(mongomock.MongoClient())

    with monkeypatch.context() as mp:
        mp.setattr("watertap.edb.db_api.MongoClient", client_factory)
        yield client_factory


@pytest.fixture(scope="function")
def edb(mock_edb) -> ElectrolyteDB:
    # NOTE to switch to a non-mock EDB instance, simply do not use "mock_edb" in this fixture
    edb = ElectrolyteDB()
    assert edb._client is mock_edb.instance
    return edb


@pytest.fixture(scope="function")
def empty_edb(edb) -> ElectrolyteDB:
    edb._client.drop_database(edb.database)
    return edb


@pytest.fixture(scope="function")
def populated_edb(empty_edb) -> ElectrolyteDB:
    edb = empty_edb
    commands._load_bootstrap(empty_edb)
    return edb


@contextlib.contextmanager
def _changing_cwd(dest: Path) -> Path:
    origin = Path.cwd()
    try:
        os.chdir(dest)
        yield dest
    finally:
        os.chdir(origin)


@pytest.fixture(scope="function")
def run_in_empty_dir(tmp_path: Path) -> Path:
    with _changing_cwd(tmp_path) as dest:
        yield dest


class TestBaseCommand:
    @pytest.fixture(
        scope="function",
        params=[
            (["-q"], Expected(log_level=logging.FATAL)),
            ([], Expected(log_level=logging.ERROR)),
            (["-v"], Expected(log_level=logging.WARN)),
            (["-vv"], Expected(log_level=logging.INFO)),
            (["-vvv"], Expected(log_level=logging.DEBUG)),
            (["-qv"], Expected(exit_code=ExitCode.INVALID_USAGE)),
        ],
        ids=_display_param,
    )
    def run_command(
        self,
        request,
        runner,
        edb,
        mandatory_subcommand_args=tuple(["info", "--type", "reaction"]),
    ) -> Outcome:
        flags, expected = request.param
        args = flags + list(mandatory_subcommand_args)
        result = runner.invoke(commands.command_base, args)
        return result, expected

    @pytest.fixture(
        scope="function",
    )
    def current_log_level(self, logger_to_check: str = "watertap") -> LogLevel:
        logger = logging.getLogger(logger_to_check)
        return logger.getEffectiveLevel()

    @pytest.mark.unit
    def test_verbosity_options_set_log_level(
        self, run_command: Outcome, current_log_level: LogLevel
    ):
        result, expected = run_command
        assert result == expected
        if expected.success:
            assert current_log_level == expected.log_level


class TestLoad:
    @pytest.fixture(scope="function")
    def access_loadable_data(self, tmp_path: Path) -> Path:
        src = commands.get_edb_data("filename").parent
        dest = tmp_path / "data"

        # NOTE: dirs_exist_ok is only available for Python 3.8+
        shutil.copytree(src, dest, dirs_exist_ok=False)
        with _changing_cwd(dest):
            yield dest

    @pytest.fixture(
        scope="function",
        params=[
            pytest.param(
                Invocation(
                    ["load", "--bootstrap", "--validate"], Expected(ExitCode.OK)
                ),
                marks=pytest.mark.xfail(
                    reason="Validation for 'base' not yet available"
                ),
            ),
            Invocation(["load", "--bootstrap", "--no-validate"]),
            Invocation(["load"], Expected(ExitCode.INVALID_USAGE)),
            Invocation(["load", "--file", "reaction.json", "--type", "reaction"]),
            Invocation(["load", "--file", "component.json", "--type", "component"]),
            Invocation(["load", "--file", "base.json", "--type", "base"]),
            Invocation(
                ["load", "--file", "base.json"], Expected(ExitCode.INVALID_USAGE)
            ),
            # to test the validation from the point of view of the command line, we use valid files and data types,
            # but switched, e.g. base.json as component.
            # the command should fail unless the validation is turned off (default is on)
            Invocation(
                ["load", "--file", "base.json", "--type", "component", "--no-validate"],
                Expected(ExitCode.OK),
            ),
            Invocation(
                ["load", "--file", "base.json", "--type", "component", "--validate"],
                Expected(ExitCode.DATABASE_ERROR),
            ),
            # this tests that the default behavior is to validate if no flag is given
            Invocation(
                ["load", "--file", "base.json", "--type", "component"],
                Expected(ExitCode.DATABASE_ERROR),
            ),
            Invocation(
                ["load", "--bootstrap", "--url", _INVALID.url],
                Expected(ExitCode.DATABASE_ERROR),
            ),
        ],
        ids=_display_param,
    )
    def run_from_empty_db(
        self, request, runner, empty_edb, access_loadable_data
    ) -> Outcome:
        invocation = request.param
        result = runner.invoke(commands.command_base, invocation.args)
        return result, invocation.expected

    @pytest.mark.unit
    def test_from_empty_db(self, run_from_empty_db: Outcome, edb: ElectrolyteDB):
        result, expected = run_from_empty_db
        assert result == expected, result.stdout
        if expected.success:
            assert not edb.is_empty()

    @pytest.fixture(
        scope="function",
        params=[
            (["load", "--bootstrap"], Expected(ExitCode.DATABASE_ERROR)),
        ],
        ids=_display_param,
    )
    def run_from_populated_db(self, request, runner, populated_edb) -> Outcome:
        args, expected = request.param
        result = runner.invoke(commands.command_base, args)
        return result, expected

    @pytest.mark.unit
    def test_from_populated_db(
        self, run_from_populated_db: Outcome, edb: ElectrolyteDB
    ):
        result, expected = run_from_populated_db
        assert result == expected, result.stdout
        assert not edb.is_empty()


class TestDump:
    @pytest.fixture(
        scope="function",
        params=[
            (
                ["dump", "--type", "reaction", "--file", "reaction.json"],
                Expected(file_name="reaction.json"),
            ),
            (
                ["dump", "--type", "base", "--file", "base.json"],
                Expected(file_name="base.json"),
            ),
            (
                ["dump", "--type", "component", "--file", "component.json"],
                Expected(file_name="component.json"),
            ),
            (
                ["dump", "--type", _INVALID.data_type, "--file", "invalid.json"],
                Expected(ExitCode.INVALID_USAGE),
            ),
        ],
        ids=_display_param,
    )
    def run_from_populated_db(
        self, request, runner, populated_edb, tmp_path
    ) -> Outcome:
        args, expected = request.param
        with _changing_cwd(tmp_path) as dest_dir:
            result = runner.invoke(commands.command_base, args)
        if expected.file_name:
            expected.file_path = dest_dir / expected.file_name
        yield result, expected

    @pytest.mark.unit
    def test_from_populated_db(self, run_from_populated_db: Outcome):
        result, expected = run_from_populated_db
        assert result == expected, result.stdout
        if expected.file_name is not None:
            assert expected.file_path.exists()


class TestDrop:
    @pytest.fixture(
        scope="function",
        params=[
            (["drop", "--yes"], Expected(ExitCode.OK)),
            # TODO handle case where expected behavior is to prompt the user,
            # so that we can test this command without the "--yes" flag as well
        ],
        ids=_display_param,
    )
    def run_from_populated_db(self, request, runner, populated_edb) -> Outcome:
        args, expected = request.param
        result = runner.invoke(commands.command_base, args)
        return result, expected

    @pytest.mark.unit
    def test_from_populated_db(self, run_from_populated_db: Outcome, edb):
        result, expected = run_from_populated_db
        assert result == expected, result.stdout
        if expected.success:
            assert edb.is_empty()


class TestInfo:
    @pytest.fixture(
        scope="function",
        params=[
            (["info", "--type", "base"], Expected(ExitCode.OK)),
            (["info", "--type", "component"], Expected(ExitCode.OK)),
            (["info", "--type", "reaction"], Expected(ExitCode.OK)),
            (["info", "--type", _INVALID.data_type], Expected(ExitCode.INVALID_USAGE)),
        ],
        ids=_display_param,
    )
    def run_command(self, request, runner, edb) -> Outcome:
        args, expected = request.param
        result = runner.invoke(commands.command_base, args)
        return result, expected

    @pytest.mark.unit
    def test_command(self, run_command: Outcome, min_text_length=10):
        result, expected = run_command
        assert result == expected, result.stdout
        if expected.success:
            assert len(result.stdout) >= min_text_length


class TestSchema:
    format_available = {
        "json": True,
        "json-compact": False,
        "markdown": False,
        "html": False,
        "html-js": False,
    }

    @pytest.fixture(
        scope="function",
        params=[
            fmt
            if is_available
            else pytest.param(
                fmt,
                marks=pytest.mark.xfail(
                    reason=f"Schema output format '{fmt}' not yet supported"
                ),
            )
            for fmt, is_available in format_available.items()
        ],
        ids=str,
    )
    def output_format(self, request):
        return request.param

    @pytest.fixture(
        scope="function",
        params=[
            (["schema", "--type", "component"], Expected(ExitCode.OK)),
            (["schema", "--type", "reaction"], Expected(ExitCode.OK)),
            (["schema", "--type", "base"], Expected(ExitCode.INVALID_USAGE)),
            (
                ["schema", "--type", "component", "--file", "component.json"],
                Expected(ExitCode.OK, file_name="component.json"),
            ),
        ],
        ids=_display_param,
    )
    def run_command(
        self, request, output_format: str, runner, edb, run_in_empty_dir
    ) -> Outcome:
        args, expected = request.param
        args.extend(["--format", output_format])
        result = runner.invoke(commands.command_base, args)
        return result, expected

    @pytest.mark.unit
    def test_command(self, run_command: Outcome, min_text_length=20):
        result, expected = run_command
        assert result == expected, result.stdout
        if expected.success:
            text = (
                expected.file_path.read_text() if expected.file_path else result.stdout
            )
            assert len(text) >= min_text_length



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pyomo.environ as pyo
from idaes.core.util.exceptions import ConfigurationError
from idaes.core.util.misc import StrEnum
from ..util import (
    cost_by_flow_volume,
    register_costing_parameter_block,
    make_capital_cost_var,
)


class PumpType(StrEnum):
    low_pressure = "low_pressure"
    high_pressure = "high_pressure"


def cost_pump(blk, pump_type=PumpType.high_pressure, cost_electricity_flow=True):
    """
    Pump costing method

    TODO: describe equations

    Args:
        pump_type: PumpType Enum indicating pump type,
            default = PumpType.high_pressure

        cost_electricity_flow: bool, if True, the Pump's work_mechanical will be
            converted to kW and costed as an electricity, default = True
    """
    if pump_type == PumpType.high_pressure:
        cost_high_pressure_pump(blk, cost_electricity_flow)
    elif pump_type == PumpType.low_pressure:
        cost_low_pressure_pump(blk, cost_electricity_flow)
    else:
        raise ConfigurationError(
            f"{blk.unit_model.name} received invalid argument for pump_type:"
            f" {pump_type}. Argument must be a member of the PumpType Enum."
        )


def build_high_pressure_pump_cost_param_block(blk):

    blk.cost = pyo.Var(
        initialize=53 / 1e5 * 3600,
        bounds=(0, None),
        doc="High pressure pump cost",
        units=pyo.units.USD_2018 / pyo.units.watt,
    )


@register_costing_parameter_block(
    build_rule=build_high_pressure_pump_cost_param_block,
    parameter_block_name="high_pressure_pump",
)
def cost_high_pressure_pump(blk, cost_electricity_flow=True):
    """
    High pressure pump costing method

    `TODO: describe equations`

    Args:
        cost_electricity_flow (bool): if True, the Pump's work_mechanical will
            be converted to kW and costed as an electricity. Defaults to True.
    """
    t0 = blk.flowsheet().time.first()
    make_capital_cost_var(blk)
    blk.costing_package.add_cost_factor(blk, "TIC")
    blk.capital_cost_constraint = pyo.Constraint(
        expr=blk.capital_cost
        == blk.cost_factor
        * pyo.units.convert(
            blk.costing_package.high_pressure_pump.cost
            * pyo.units.convert(blk.unit_model.work_mechanical[t0], pyo.units.W),
            to_units=blk.costing_package.base_currency,
        )
    )
    if cost_electricity_flow:
        # grab lower bound of mechanical work
        lb = blk.unit_model.work_mechanical[t0].lb
        # set lower bound to 0 to avoid negative defined flow warning when lb is not >= 0
        blk.unit_model.work_mechanical.setlb(0)
        blk.costing_package.cost_flow(
            pyo.units.convert(
                blk.unit_model.work_mechanical[t0], to_units=pyo.units.kW
            ),
            "electricity",
        )
        # set lower bound back to its original value that was assigned to lb
        blk.unit_model.work_mechanical.setlb(lb)


def build_low_pressure_pump_cost_param_block(blk):

    blk.cost = pyo.Var(
        initialize=889,
        doc="Low pressure pump cost",
        units=pyo.units.USD_2018 / (pyo.units.liter / pyo.units.second),
    )


@register_costing_parameter_block(
    build_rule=build_low_pressure_pump_cost_param_block,
    parameter_block_name="low_pressure_pump",
)
def cost_low_pressure_pump(blk, cost_electricity_flow=True):
    """
    Low pressure pump costing method

    TODO: describe equations

    Args:
        cost_electricity_flow (bool): if True, the Pump's work_mechanical will
            be converted to kW and costed as an electricity. Defaults to True.
    """
    t0 = blk.flowsheet().time.first()
    cost_by_flow_volume(
        blk,
        blk.costing_package.low_pressure_pump.cost,
        pyo.units.convert(
            blk.unit_model.control_volume.properties_in[t0].flow_vol,
            (pyo.units.m**3 / pyo.units.s),
        ),
    )
    if cost_electricity_flow:
        blk.costing_package.cost_flow(
            pyo.units.convert(
                blk.unit_model.work_mechanical[t0], to_units=pyo.units.kW
            ),
            "electricity",
        )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
[1] Sharma, Jwala R., Mohammad Najafi, and Syed R. Qasim.
"Preliminary cost estimation models for construction, operation, and maintenance of water treatment plants."
Journal of Infrastructure Systems 19.4 (2013): 451-464.

[2] Byun, Jaewon, Maravelias, Christos.
Benchmark Model for Wastewater Treatment Using an Activated Sludge Process.
United States: N.p., 21 Jan, 2022. Web. doi: 10.7481/1844539.
"""
import pyomo.environ as pyo
from idaes.core.util.exceptions import ConfigurationError
from idaes.core.util.misc import StrEnum
from ..util import (
    make_capital_cost_var,
    register_costing_parameter_block,
)


class ClarifierType(StrEnum):
    circular = "circular"
    rectangular = "rectangular"
    primary = "primary"


def cost_clarifier(blk, clarifier_type=ClarifierType.circular, **kwargs):
    """
    Clarifier costing method

    Args:
        clarifier_type: ClarifierType Enum indicating clarifier type,
            default = ClarifierType.circular
    """
    if clarifier_type == clarifier_type.circular:
        cost_circular_clarifier(blk, **kwargs)
    elif clarifier_type == clarifier_type.rectangular:
        cost_rectangular_clarifier(blk, **kwargs)
    elif clarifier_type == clarifier_type.primary:
        cost_primary_clarifier(blk, **kwargs)
    else:
        raise ConfigurationError(
            f"{blk.unit_model.name} received invalid argument for clarifier_type:"
            f" {clarifier_type}. Argument must be a member of the ClarifierType Enum."
        )


def build_circular_clarifier_cost_param_block(blk):

    blk.construction_a_parameter = pyo.Var(
        initialize=-6e-4,
        doc="A parameter for construction cost",
        units=pyo.units.USD_2011 / pyo.units.ft**4,
    )

    blk.construction_b_parameter = pyo.Var(
        initialize=98.952,
        doc="B parameter for construction cost",
        units=pyo.units.USD_2011 / pyo.units.ft**2,
    )

    blk.construction_c_parameter = pyo.Var(
        initialize=191806,
        doc="C parameter for construction cost",
        units=pyo.units.USD_2011,
    )


@register_costing_parameter_block(
    build_rule=build_circular_clarifier_cost_param_block,
    parameter_block_name="circular",
)
def cost_circular_clarifier(blk, cost_electricity_flow=True):
    """
    Circular clarifier costing method [1]
    """
    make_capital_cost_var(blk)
    blk.costing_package.add_cost_factor(blk, "TIC")

    surface_area = pyo.units.convert(
        blk.unit_model.surface_area, to_units=pyo.units.ft**2
    )

    blk.capital_cost_constraint = pyo.Constraint(
        expr=blk.capital_cost
        == blk.cost_factor
        * pyo.units.convert(
            blk.costing_package.circular.construction_a_parameter * surface_area**2
            + blk.costing_package.circular.construction_b_parameter * surface_area
            + blk.costing_package.circular.construction_c_parameter,
            to_units=blk.costing_package.base_currency,
        )
    )

    t0 = blk.flowsheet().time.first()
    if cost_electricity_flow:
        blk.costing_package.cost_flow(
            pyo.units.convert(
                blk.unit_model.electricity_consumption[t0],
                to_units=pyo.units.kW,
            ),
            "electricity",
        )


def build_rectangular_clarifier_cost_param_block(blk):

    blk.construction_a_parameter = pyo.Var(
        initialize=-2.9e-3,
        doc="A parameter for construction cost",
        units=pyo.units.USD_2011 / pyo.units.ft**4,
    )

    blk.construction_b_parameter = pyo.Var(
        initialize=169.19,
        doc="B parameter for construction cost",
        units=pyo.units.USD_2011 / pyo.units.ft**2,
    )

    blk.construction_c_parameter = pyo.Var(
        initialize=94365,
        doc="C parameter for construction cost",
        units=pyo.units.USD_2011,
    )


@register_costing_parameter_block(
    build_rule=build_rectangular_clarifier_cost_param_block,
    parameter_block_name="rectangular",
)
def cost_rectangular_clarifier(blk, cost_electricity_flow=True):
    """
    Rectangular clarifier costing method [1]
    """
    make_capital_cost_var(blk)
    blk.costing_package.add_cost_factor(blk, "TIC")

    surface_area = pyo.units.convert(
        blk.unit_model.surface_area, to_units=pyo.units.ft**2
    )

    blk.capital_cost_constraint = pyo.Constraint(
        expr=blk.capital_cost
        == blk.cost_factor
        * pyo.units.convert(
            blk.costing_package.rectangular.construction_a_parameter * surface_area**2
            + blk.costing_package.rectangular.construction_b_parameter * surface_area
            + blk.costing_package.rectangular.construction_c_parameter,
            to_units=blk.costing_package.base_currency,
        )
    )

    t0 = blk.flowsheet().time.first()
    if cost_electricity_flow:
        blk.costing_package.cost_flow(
            pyo.units.convert(
                blk.unit_model.electricity_consumption[t0],
                to_units=pyo.units.kW,
            ),
            "electricity",
        )


def build_primary_clarifier_cost_param_block(blk):

    blk.capital_a_parameter = pyo.Var(
        initialize=120000 / 2776 * 12463,
        doc="A parameter for capital cost",
        units=pyo.units.USD_2021,
    )

    blk.capital_b_parameter = pyo.Var(
        initialize=0.7,
        doc="B parameter for construction cost",
        units=pyo.units.dimensionless,
    )


@register_costing_parameter_block(
    build_rule=build_primary_clarifier_cost_param_block,
    parameter_block_name="primary",
)
def cost_primary_clarifier(blk, cost_electricity_flow=True):
    """
    Primary clarifier costing method [2]
    """
    make_capital_cost_var(blk)
    blk.costing_package.add_cost_factor(blk, "TIC")

    t0 = blk.flowsheet().time.first()
    flow_in = pyo.units.convert(
        blk.unit_model.inlet.flow_vol[t0], to_units=pyo.units.gallon / pyo.units.day
    )
    blk.capital_cost_constraint = pyo.Constraint(
        expr=blk.capital_cost
        == blk.cost_factor
        * pyo.units.convert(
            blk.costing_package.primary.capital_a_parameter
            * pyo.units.convert(
                flow_in / (1e6 * pyo.units.gallon / pyo.units.day),
                to_units=pyo.units.dimensionless,
            )
            ** blk.costing_package.primary.capital_b_parameter,
            to_units=blk.costing_package.base_currency,
        )
    )

    if cost_electricity_flow:
        blk.costing_package.cost_flow(
            pyo.units.convert(
                blk.unit_model.electricity_consumption[t0],
                to_units=pyo.units.kW,
            ),
            "electricity",
        )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

"""Seawater feed specifications for supported property packages"""

from pyomo.environ import ConcreteModel
from idaes.core import FlowsheetBlock
from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterBlock,
)
from idaes.core.util.scaling import calculate_scaling_factors
from watertap.property_models import seawater_prop_pack
from watertap.property_models import seawater_ion_prop_pack
from watertap.examples.flowsheets.full_treatment_train.model_components import (
    seawater_salt_prop_pack,
)
from watertap.examples.flowsheets.full_treatment_train.model_components.eNRTL import (
    entrl_config_FpcTP,
)
from watertap.examples.flowsheets.full_treatment_train.util import solve_block


def build_prop(m, base="TDS"):
    """
    Builds a property package for the specified base. Includes default scaling.
    Bases include: 'TDS', 'ion', 'salt'.
    """
    if base == "TDS":
        m.fs.prop_TDS = seawater_prop_pack.SeawaterParameterBlock()

        m.fs.prop_TDS.set_default_scaling(
            "flow_mass_phase_comp", 1, index=("Liq", "H2O")
        )
        m.fs.prop_TDS.set_default_scaling(
            "flow_mass_phase_comp", 1e2, index=("Liq", "TDS")
        )

    elif base == "ion":
        m.fs.prop_ion = seawater_ion_prop_pack.PropParameterBlock()

        m.fs.prop_ion.set_default_scaling(
            "flow_mass_phase_comp", 1, index=("Liq", "H2O")
        )
        m.fs.prop_ion.set_default_scaling(
            "flow_mass_phase_comp", 1e2, index=("Liq", "Na")
        )
        m.fs.prop_ion.set_default_scaling(
            "flow_mass_phase_comp", 1e4, index=("Liq", "Ca")
        )
        m.fs.prop_ion.set_default_scaling(
            "flow_mass_phase_comp", 1e3, index=("Liq", "Mg")
        )
        m.fs.prop_ion.set_default_scaling(
            "flow_mass_phase_comp", 1e3, index=("Liq", "SO4")
        )
        m.fs.prop_ion.set_default_scaling(
            "flow_mass_phase_comp", 1e2, index=("Liq", "Cl")
        )

    elif base == "salt":
        m.fs.prop_salt = seawater_salt_prop_pack.PropParameterBlock()

        m.fs.prop_salt.set_default_scaling(
            "flow_mass_phase_comp", 1, index=("Liq", "H2O")
        )
        m.fs.prop_salt.set_default_scaling(
            "flow_mass_phase_comp", 1e2, index=("Liq", "NaCl")
        )
        m.fs.prop_salt.set_default_scaling(
            "flow_mass_phase_comp", 1e3, index=("Liq", "CaSO4")
        )
        m.fs.prop_salt.set_default_scaling(
            "flow_mass_phase_comp", 1e3, index=("Liq", "MgSO4")
        )
        m.fs.prop_salt.set_default_scaling(
            "flow_mass_phase_comp", 1e3, index=("Liq", "MgCl2")
        )

    elif base == "eNRTL":
        m.fs.prop_eNRTL = GenericParameterBlock(**entrl_config_FpcTP.configuration)

        # default scaling in config file

    else:
        raise ValueError(
            "Unexpected property base {base} provided to build_prop"
            "".format(base=base)
        )


def get_prop(m, base="TDS"):
    if base == "TDS":
        prop = m.fs.prop_TDS
    elif base == "ion":
        prop = m.fs.prop_ion
    elif base == "salt":
        prop = m.fs.prop_salt
    elif base == "eNRTL":
        prop = m.fs.prop_eNRTL
    else:
        raise ValueError(
            "Unexpected property base {base} for get_prop" "".format(base=base)
        )
    return prop


def specify_feed(sb, base="TDS"):
    """
    Fixes the state variables on the stateblock to the base seawater composition for
    the specified base. Bases include: 'TDS', 'ion', 'salt'.
    """
    sb.pressure.fix(101325)
    sb.temperature.fix(298.15)

    feed_flow_mass = 1
    if base == "TDS":
        feed_mass_frac_TDS = 0.035
        sb.flow_mass_phase_comp["Liq", "TDS"].fix(feed_flow_mass * feed_mass_frac_TDS)
        sb.flow_mass_phase_comp["Liq", "H2O"].fix(
            feed_flow_mass * (1 - feed_mass_frac_TDS)
        )

    elif base == "ion":
        feed_mass_frac = {
            "Na": 11122e-6,
            "Ca": 382e-6,
            "Mg": 1394e-6,
            "SO4": 2136e-6,
            "Cl": 20316.88e-6,
        }
        sb.flow_mass_phase_comp["Liq", "H2O"].fix(
            feed_flow_mass * (1 - sum(x for x in feed_mass_frac.values()))
        )
        for j in feed_mass_frac:
            sb.flow_mass_phase_comp["Liq", j].fix(feed_flow_mass * feed_mass_frac[j])

    elif base == "salt":
        feed_mass_frac = {
            "NaCl": 2.827e-2,
            "CaSO4": 1.298e-3,
            "MgSO4": 1.529e-3,
            "MgCl2": 4.251e-3,
            "H2O": 0.9647,
        }
        for s in feed_mass_frac:
            sb.flow_mass_phase_comp["Liq", s].fix(feed_flow_mass * feed_mass_frac[s])

    else:
        raise ValueError(
            "Unexpected property base {base} provided to specify_feed"
            "".format(base=base)
        )


def solve_specify_feed(base):
    # build state block
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    build_prop(m, base=base)
    if base == "TDS":
        m.fs.stream = m.fs.prop_TDS.build_state_block([0])
    elif base == "ion":
        m.fs.stream = m.fs.prop_ion.build_state_block([0])
    elif base == "salt":
        m.fs.stream = m.fs.prop_salt.build_state_block([0])
    specify_feed(m.fs.stream[0], base=base)

    m.fs.stream[
        0
    ].mass_frac_phase_comp  # touch a variable to have a model with at least one constraint

    # scale
    calculate_scaling_factors(m.fs)
    # solve
    solve_block(m)
    # display
    m.fs.stream.display()

    return m


if __name__ == "__main__":
    solve_specify_feed("TDS")
    solve_specify_feed("ion")
    solve_specify_feed("salt")
